"""Data transfer objects for Matchbox API."""

import json
import re
import textwrap
from collections.abc import Iterable
from enum import StrEnum
from importlib.metadata import version
from typing import Annotated, Any, Self, TypeAlias

from pydantic import (
    BaseModel,
    ConfigDict,
    EmailStr,
    Field,
    GetCoreSchemaHandler,
    GetJsonSchemaHandler,
    PlainSerializer,
    PlainValidator,
    field_serializer,
    field_validator,
    model_validator,
)
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema

from matchbox.common.datatypes import DataTypes
from matchbox.common.exceptions import MatchboxExceptionType, MatchboxNameError
from matchbox.common.hash import base64_to_hash, hash_to_base64


class MatchboxName(str):
    """Sub-class of string which validates names for the Matchbox DB."""

    PATTERN = r"^[a-zA-Z0-9_.-]+$"

    def __new__(cls, value: str) -> Self:
        """Creates new instance of validated name."""
        if not isinstance(value, str):
            raise MatchboxNameError("Name must be a string")
        if not re.match(cls.PATTERN, value):
            raise MatchboxNameError(
                f"Name '{value}' is invalid. It can only include "
                "alphanumeric characters, underscores, dots or hyphens."
            )

        return super().__new__(cls, value)

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        """Generate core schema for Pydantic compatibility."""
        return core_schema.no_info_plain_validator_function(
            cls,
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda x: x, return_schema=core_schema.str_schema()
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        schema: core_schema.CoreSchema,
        handler: GetJsonSchemaHandler,
    ) -> JsonSchemaValue:
        """Generate JSON schema for OpenAPI documentation."""
        json_schema = handler(core_schema.str_schema())
        json_schema.update(
            {
                "type": "string",
                "pattern": cls.PATTERN,
            }
        )
        return json_schema


class OKMessage(BaseModel):
    """Generic HTTP OK response."""

    status: str = Field(default="OK")
    version: str = Field(default_factory=lambda: version("matchbox-db"))


class BackendCountableType(StrEnum):
    """Enumeration of supported backend countable types."""

    SOURCES = "sources"
    MODELS = "models"
    SOURCE_CLUSTERS = "source_clusters"
    MODEL_CLUSTERS = "model_clusters"
    CLUSTERS = "all_clusters"
    CREATES = "creates"
    MERGES = "merges"
    PROPOSES = "proposes"


class ModelResultsType(StrEnum):
    """Enumeration of supported model results types."""

    PROBABILITIES = "probabilities"
    CLUSTERS = "clusters"


class BackendResourceType(StrEnum):
    """Enumeration of resources types referenced by client or API."""

    COLLECTION = "collection"
    RUN = "run"
    RESOLUTION = "resolution"
    CLUSTER = "cluster"
    USER = "user"
    JUDGEMENT = "judgement"
    SYSTEM = "system"


class BackendParameterType(StrEnum):
    """Enumeration of parameter types passable to the API."""

    SAMPLE_SIZE = "sample_size"
    NAME = "name"


class CRUDOperation(StrEnum):
    """Enumeration of CRUD operations."""

    CREATE = "create"
    UPDATE = "update"
    DELETE = "delete"


class LocationType(StrEnum):
    """Enumeration of location types."""

    RDBMS = "rdbms"


class PermissionType(StrEnum):
    """Permission levels for resource access."""

    READ = "read"
    WRITE = "write"
    ADMIN = "admin"


class User(BaseModel):
    """User identity."""

    model_config = ConfigDict(populate_by_name=True)

    user_name: str = Field(alias="sub")
    email: EmailStr | None = None


GroupName: TypeAlias = MatchboxName
"""Type alias for group names."""


class Group(BaseModel):
    """Group definition."""

    name: GroupName
    description: str | None = None
    is_system: bool = False
    members: list[User] = []


class PermissionGrant(BaseModel):
    """A permission on a resource.

    Resource context should always be supplied.
    """

    group_name: GroupName
    permission: PermissionType


class AuthStatusResponse(BaseModel):
    """Response model for authentication status."""

    authenticated: bool
    username: str | None = None
    token: str | None = None


CollectionName: TypeAlias = MatchboxName
"""Type alias for collection names."""

RunID: TypeAlias = int
"""Type alias for run IDs."""

SourceResolutionName: TypeAlias = MatchboxName
"""Type alias for source resolution names."""

ModelResolutionName: TypeAlias = MatchboxName
"""Type alias for model resolution names."""

ResolutionName: TypeAlias = SourceResolutionName | ModelResolutionName
"""Type alias for any resolution names."""


class ResolutionPath(BaseModel):
    """Base resolution identifier with collection, run, and name."""

    model_config = ConfigDict(frozen=True)

    collection: CollectionName
    run: RunID
    name: ResolutionName

    def __str__(self) -> str:
        """String representation of the resolution path."""
        return f"{self.collection}/{self.run}/{self.name}"


SourceResolutionPath: TypeAlias = ResolutionPath
"""Type alias for source resolution paths."""

ModelResolutionPath: TypeAlias = ResolutionPath
"""Type alias for model resolution paths."""


class ResolutionType(StrEnum):
    """Types of nodes in a resolution."""

    SOURCE = "source"
    MODEL = "model"


class LocationConfig(BaseModel):
    """Metadata for a location."""

    model_config = ConfigDict(frozen=True)

    type: LocationType
    name: str


class SourceField(BaseModel):
    """A field in a source that can be indexed in the Matchbox database."""

    model_config = ConfigDict(frozen=True)

    name: str = Field(
        description=(
            "The name of the field in the source after the "
            "extract/transform logic has been applied."
        )
    )
    type: DataTypes = Field(
        description="The cached field type. Used to ensure a stable hash.",
    )


class SourceConfig(BaseModel):
    """Configuration of a source that can, or has been, indexed in the backend.

    They are foundational processes on top of which linking and deduplication models can
    build new resolutions.
    """

    model_config = ConfigDict(frozen=True)

    location_config: LocationConfig = Field(
        description=(
            "The location of the source. Used to run the extract/tansform logic."
        ),
    )
    extract_transform: str = Field(
        description=(
            "Logic to extract and transform data from the source. "
            "Language is location dependent."
        )
    )
    # Fields can to be set at creation, or initialised with `.default_columns()`
    key_field: SourceField = Field(
        description=textwrap.dedent("""
            The key field. This is the source's key for unique
            entities, such as a primary key in a relational database.

            Keys must ALWAYS be a string.

            For example, if the source describes companies, it may have used
            a Companies House number as its key.

            This key is ALWAYS correct. It should be something generated and
            owned by the source being indexed.
            
            For example, your organisation's CRM ID is a key field within the CRM.
            
            A CRM ID entered by hand in another dataset shouldn't be used 
            as a key field.
        """),
    )
    index_fields: tuple[SourceField, ...] = Field(
        default=None,
        description=textwrap.dedent(
            """
            The fields to index in this source, after the extract/transform logic 
            has been applied. 

            This is usually set manually, and should map onto the columns that the
            extract/transform logic returns.
            """
        ),
    )

    @model_validator(mode="after")
    def validate_key_field(self) -> Self:
        """Ensure that the key field is a string and not in the index fields."""
        if self.key_field in self.index_fields:
            raise ValueError("Key field must not be in the index fields. ")

        if self.key_field.type != DataTypes.STRING:
            raise ValueError("Key field must have string type.")

        return self

    @property
    def dependencies(self) -> list[ResolutionName]:
        """Return all resolution names that this source needs.

        Provided for symmetry with ModelConfig.
        """
        return []

    @property
    def parents(self) -> list[ResolutionName]:
        """Returns all resolution names directly input to this config.

        Provided for symmetry with ModelConfig.
        """
        return []

    def prefix(self, name: str) -> str:
        """Get the prefix for the source.

        Args:
            name: The name of the source.

        Returns:
            The prefix string (name + "_").
        """
        return name + "_"

    def qualified_key(self, name: str) -> str:
        """Get the qualified key for the source.

        Args:
            name: The name of the source.

        Returns:
            The qualified key field name.
        """
        return self.qualify_field(name, self.key_field.name)

    def qualified_index_fields(self, name: str) -> list[str]:
        """Get the qualified index fields for the source.

        Args:
            name: The name of the source.

        Returns:
            List of qualified index field names.
        """
        return [self.qualify_field(name, field.name) for field in self.index_fields]

    def qualify_field(self, name: str, field: str) -> str:
        """Qualify field names with the source name.

        Args:
            name: The name of the source.
            field: The field name to qualify.

        Returns:
            A single qualified field.
        """
        return self.prefix(name) + field

    def f(self, name: str, fields: str | Iterable[str]) -> str | list[str]:
        """Qualify one or more field names with the source name.

        Args:
            name: The name of the source.
            fields: The field name to qualify, or a list of field names.

        Returns:
            A single qualified field, or a list of qualified field names.
        """
        if isinstance(fields, str):
            return self.qualify_field(name, fields)
        return [self.qualify_field(name, field_name) for field_name in fields]


class QueryCombineType(StrEnum):
    """Enumeration of ways to combine multiple rows having the same matchbox ID."""

    CONCAT = "concat"
    EXPLODE = "explode"
    SET_AGG = "set_agg"


class QueryConfig(BaseModel):
    """Configuration of query generating model inputs."""

    model_config = ConfigDict(frozen=True)

    source_resolutions: tuple[SourceResolutionName, ...]
    model_resolution: ModelResolutionName | None = None
    combine_type: QueryCombineType = QueryCombineType.CONCAT
    threshold: int | None = None
    cleaning: dict[str, str] | None = None

    @model_validator(mode="after")
    def validate_resolutions(self) -> Self:
        """Ensure that resolution settings are compatible."""
        if not self.source_resolutions:
            raise ValueError("At least one source resolution required.")
        if len(self.source_resolutions) > 1 and not self.model_resolution:
            raise ValueError(
                "A model resolution must be set if querying from multiple sources"
            )
        return self

    @property
    def dependencies(self) -> list[ResolutionName]:
        """Return all resolutions that this query needs."""
        deps = list(self.source_resolutions)
        if self.model_resolution:
            deps.append(self.model_resolution)

        return deps

    @property
    def point_of_truth(self) -> ResolutionName:
        """Return path of resolution that will be used as point of truth."""
        if self.model_resolution:
            return self.model_resolution
        return self.source_resolutions[0]


class ModelType(StrEnum):
    """Enumeration of supported model types."""

    LINKER = "linker"
    DEDUPER = "deduper"


class ModelConfig(BaseModel):
    """Configuration for model that has or could be added to the server."""

    type: ModelType
    model_class: str
    model_settings: str
    left_query: QueryConfig
    right_query: QueryConfig | None = None  # Only used for linker models

    def __eq__(self, other: object) -> bool:
        """Check equality of model configurations.

        Model configurations don't care about the order of left and right resolutions.
        """
        if not isinstance(other, ModelConfig):
            return NotImplemented
        return self.type == other.type and {
            self.left_query,
            self.right_query,
        } == {other.left_query, other.right_query}

    @model_validator(mode="after")
    def validate_right_query(self) -> Self:
        """Ensure that a right query is set if and only if model is linker."""
        if self.type == ModelType.DEDUPER and self.right_query is not None:
            raise ValueError("Right query can't be set for dedupers")
        if self.type == ModelType.LINKER and self.right_query is None:
            raise ValueError("Right query must be set for linkers")

        return self

    @field_validator("model_settings", mode="after")
    @classmethod
    def validate_settings_json(cls, value: str) -> str:
        """Ensure that the model settings is valid JSON."""
        try:
            isinstance(json.loads(value), dict)
        except json.JSONDecodeError as e:
            raise ValueError("Model settings are not valid JSON") from e
        return value

    @property
    def dependencies(self) -> list[ResolutionName]:
        """Return all resolutions that this model needs."""
        deps = list(self.left_query.dependencies)
        if self.right_query:
            deps.extend(self.right_query.dependencies)

        return deps

    @property
    def parents(self) -> list[ResolutionName]:
        """Returns all resolution names directly input to this config."""
        if self.right_query:
            return [
                self.left_query.point_of_truth,
                self.right_query.point_of_truth,
            ]
        return [self.left_query.point_of_truth]


class Match(BaseModel):
    """A match between primary keys in the Matchbox database."""

    cluster: int | None
    source: SourceResolutionPath
    source_id: set[str] = Field(default_factory=set)
    target: SourceResolutionPath
    target_id: set[str] = Field(default_factory=set)

    @model_validator(mode="after")
    def found_or_none(self) -> "Match":
        """Ensure that a match has sources and a cluster if target was found."""
        if self.target_id and not (self.source_id and self.cluster):
            raise ValueError(
                "A match must have sources and a cluster if target was found."
            )
        if self.cluster and not self.source_id:
            raise ValueError("A match must have source if cluster is set.")
        return self

    @field_serializer("source_id", "target_id")
    def serialise_ids(self, id_set: set[str]) -> list[str]:
        """Turn set to sorted list when serialising."""
        return sorted(id_set)


class Resolution(BaseModel):
    """Unified resolution type with common fields and discriminated config."""

    description: str | None = Field(default=None, description="Description")
    truth: int | None = Field(default=None, ge=0, le=100, strict=True)
    fingerprint: Annotated[
        bytes,
        PlainSerializer(hash_to_base64, return_type=str),
        PlainValidator(base64_to_hash),
    ]

    # Discriminator field
    resolution_type: ResolutionType

    # Type-specific config as discriminated union
    config: SourceConfig | ModelConfig

    @field_validator("description", mode="after")
    @classmethod
    def validate_description(cls, value: str | None) -> str | None:
        """Ensure the description is not empty if provided."""
        if value is not None and not value.strip():
            raise ValueError("Description cannot be empty if provided.")
        return value

    @model_validator(mode="after")
    def validate_resolution_type_matches_config(self) -> Self:
        """Ensure resolution_type matches the config type."""
        if self.resolution_type == ResolutionType.SOURCE:
            assert isinstance(self.config, SourceConfig), (
                "Config must be SourceConfig when resolution_type is 'source'"
            )
        else:
            assert isinstance(self.config, ModelConfig), (
                "Config must be ModelConfig when resolution_type is 'model'"
            )
        return self

    @model_validator(mode="after")
    def validate_truth_matches_type(self) -> Self:
        """Ensure truth field matches resolution type requirements."""
        if self.resolution_type == ResolutionType.SOURCE and self.truth is not None:
            raise ValueError("Truth must be None for source resolutions")
        elif self.resolution_type == ResolutionType.MODEL:
            if self.truth is None:
                raise ValueError("Truth is required for model resolutions")
            if not (0 <= self.truth <= 100):
                raise ValueError(
                    "Truth must be between 0 and 100 for model resolutions"
                )
        return self


class Run(BaseModel):
    """A run within a collection."""

    run_id: RunID | None = Field(description="Unique ID of the run")
    is_default: bool = Field(
        default=False,
        description="Whether this run is the default in its collection",
    )
    is_mutable: bool = Field(
        default=False, description="Whether this run can be modified"
    )
    resolutions: dict[ResolutionName, Resolution] = Field(
        default_factory=dict,
        description="Dict of resolution objects by name within this run",
    )


class Collection(BaseModel):
    """A collection of runs."""

    default_run: RunID | None = Field(
        default=None, description="ID of default run for this collection"
    )
    runs: list[RunID] = Field(
        default_factory=list, description="List of run IDs in this collection"
    )

    @model_validator(mode="after")
    def validate_default_run(self) -> Self:
        """Check default run is within all runs."""
        if self.default_run and self.default_run not in self.runs:
            raise ValueError(
                "The default run needs to be included in the list of all runs."
            )

        return self


class ResourceOperationStatus(BaseModel):
    """Status response for any resource operation."""

    success: bool
    target: str
    operation: CRUDOperation
    details: str | None = None

    @classmethod
    def error_examples(cls) -> dict:
        """Examples for error codes."""
        return {
            "content": {
                "application/json": {
                    "examples": {
                        "confirm_delete": {
                            "summary": "Delete operation requires confirmation. ",
                            "value": cls(
                                success=False,
                                target=str(
                                    ModelResolutionPath(
                                        collection="default",
                                        run=1,
                                        name="example_model",
                                    )
                                ),
                                operation=CRUDOperation.DELETE,
                                details=(
                                    "This operation will delete the resolutions "
                                    "deduper_1, deduper_2, linker_1, "
                                    "as well as all probabilities they have created. "
                                    "\n\n"
                                    "It won't delete validation associated with these "
                                    "probabilities. \n\n"
                                    "If you're sure you want to continue, rerun with "
                                    "certain=True"
                                ),
                            ).model_dump(),
                        },
                        "unhandled": {
                            "summary": (
                                "Unhandled exception encountered while updating the "
                                "model's truth value."
                            ),
                            "value": cls(
                                success=False,
                                target=str(
                                    ModelResolutionPath(
                                        collection="default",
                                        run=1,
                                        name="example_model",
                                    )
                                ),
                                operation=CRUDOperation.UPDATE,
                            ).model_dump(),
                        },
                    },
                }
            }
        }


class CountResult(BaseModel):
    """Response model for count results."""

    entities: dict[BackendCountableType, int]


class UploadStage(StrEnum):
    """Enumeration of stages of a file upload and its processing."""

    READY = "ready"
    PROCESSING = "processing"
    COMPLETE = "complete"


class UploadInfo(BaseModel):
    """Response model for file upload processes."""

    stage: UploadStage | None = None
    error: str | None = None


class NotFoundError(BaseModel):
    """API error for a 404 status code."""

    details: str
    entity: BackendResourceType


class InvalidParameterError(BaseModel):
    """API error for a custom 422 status code."""

    details: str
    parameter: BackendParameterType | None


class ErrorResponse(BaseModel):
    """Unified error response for all HTTP error status codes.

    This DTO enables the client to reconstruct the exact exception
    type that was raised on the server.
    """

    exception_type: MatchboxExceptionType = Field(
        description="The name of the exception class raised on the server"
    )
    message: str = Field(description="Human-readable error message")
    details: dict[str, Any] | None = Field(
        default=None, description="Exception-specific data for reconstruction"
    )



"""Adds Collections and Runs to scope Resolutions.

Revision ID: f500f7d832fe
Revises: c4cb937d00f4
Create Date: 2025-09-22 07:54:35.453669

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "f500f7d832fe"
down_revision: str | None = "c4cb937d00f4"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    # Create collections table
    op.create_table(
        "collections",
        sa.Column("collection_id", sa.BIGINT(), autoincrement=True, nullable=False),
        sa.Column("name", sa.TEXT(), nullable=False),
        sa.PrimaryKeyConstraint("collection_id"),
        sa.UniqueConstraint("name", name="collections_name_key"),
        schema="mb",
    )

    # Create runs table
    op.create_table(
        "runs",
        sa.Column("run_id", sa.BIGINT(), autoincrement=True, nullable=False),
        sa.Column("collection_id", sa.BIGINT(), nullable=False),
        sa.Column("is_mutable", sa.BOOLEAN(), nullable=True),
        sa.Column("is_default", sa.BOOLEAN(), nullable=True),
        sa.ForeignKeyConstraint(
            ["collection_id"], ["mb.collections.collection_id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("run_id"),
        sa.UniqueConstraint("collection_id", "run_id", name="unique_run_id"),
        schema="mb",
    )

    # Create the partial index for default runs
    op.create_index(
        "ix_default_run_collection",
        "runs",
        ["collection_id"],
        unique=True,
        schema="mb",
        postgresql_where=sa.text("is_default = true"),
    )

    # Add run_id column to resolutions (initially nullable for data migration)
    op.add_column(
        "resolutions", sa.Column("run_id", sa.BIGINT(), nullable=True), schema="mb"
    )

    # Data migration: Create default collection and run for existing resolutions
    # Only do this if there are existing resolutions
    bind = op.get_bind()
    existing_resolutions = bind.execute(
        sa.text("SELECT COUNT(*) FROM mb.resolutions")
    ).scalar()

    if existing_resolutions > 0:
        # Create default collection
        collection_result = bind.execute(
            sa.text(
                "INSERT INTO mb.collections (name) VALUES ('default') "
                "RETURNING collection_id"
            )
        )
        collection_id = collection_result.scalar()

        # Create default run
        run_result = bind.execute(
            sa.text("""
            INSERT INTO mb.runs (collection_id, is_mutable, is_default)
            VALUES (:collection_id, false, true)
            RETURNING run_id
            """),
            {"collection_id": collection_id},
        )
        run_id = run_result.scalar()

        # Associate all existing resolutions with the new default run
        bind.execute(
            sa.text("""
            UPDATE mb.resolutions
            SET run_id = :run_id
            WHERE run_id IS NULL
            """),
            {"run_id": run_id},
        )

    # Now make run_id non-nullable
    op.alter_column(
        "resolutions",
        "run_id",
        existing_type=sa.BIGINT(),
        nullable=False,
        schema="mb",
    )

    # Update unique constraints on resolutions
    op.drop_constraint(
        op.f("resolutions_name_key"), "resolutions", schema="mb", type_="unique"
    )
    op.create_unique_constraint(
        "resolutions_name_key", "resolutions", ["run_id", "name"], schema="mb"
    )

    # Create foreign key constraint for run_id
    op.create_foreign_key(
        "resolutions_run_fkey",
        "resolutions",
        "runs",
        ["run_id"],
        ["run_id"],
        source_schema="mb",
        referent_schema="mb",
        ondelete="CASCADE",
    )


def downgrade() -> None:
    """Downgrade schema."""
    # Drop foreign key constraint
    op.drop_constraint(
        "resolutions_run_fkey", "resolutions", schema="mb", type_="foreignkey"
    )

    # Revert unique constraint changes
    op.drop_constraint(
        "resolutions_name_key", "resolutions", schema="mb", type_="unique"
    )
    op.create_unique_constraint(
        op.f("resolutions_name_key"),
        "resolutions",
        ["name"],
        schema="mb",
        postgresql_nulls_not_distinct=False,
    )

    # Drop the run_id column
    op.drop_column("resolutions", "run_id", schema="mb")

    # Drop the partial index
    op.drop_index(
        "ix_default_run_collection",
        table_name="runs",
        schema="mb",
        postgresql_where=sa.text("is_default = true"),
    )

    # Drop the tables
    op.drop_table("runs", schema="mb")
    op.drop_table("collections", schema="mb")



"""Authentication routes for the Matchbox server."""

import json
from typing import Annotated

from fastapi import APIRouter, Security
from fastapi.exceptions import HTTPException

from matchbox.common.dtos import AuthStatusResponse, User
from matchbox.server.api.dependencies import (
    JWT_HEADER,
    BackendDependency,
    SettingsDependency,
    b64_decode,
    validate_jwt,
)

router = APIRouter(prefix="/auth", tags=["authentication"])


def get_username_from_token(token: str | None) -> str | None:
    """Extract username from JWT token payload.

    Args:
        token: The JWT token string.

    Returns:
        The username from the 'sub' claim, or None if token is invalid.
    """
    if not token:
        return None

    try:
        parts = token.encode().split(b".")
        if len(parts) != 3:
            return None

        _, payload_b64, _ = parts
        payload = json.loads(b64_decode(payload_b64))
        return payload.get("sub")
    except Exception as _:  # noqa: BLE001
        return None


@router.post("/login")
def login(
    backend: BackendDependency,
    credentials: User,
) -> User:
    """Receive a User with a username and returns it with a user ID."""
    return backend.login(credentials)


@router.get("/status")
async def authentication_status(
    settings: SettingsDependency,
    token: Annotated[str | None, Security(JWT_HEADER)] = None,
) -> AuthStatusResponse:
    """Check authentication status and return user details.

    Returns the current authentication status, including the username
    and JWT token if the user is authenticated.
    """
    if not settings.authorisation:
        return AuthStatusResponse(
            authenticated=False,
            username=None,
            token=None,
        )

    # Try to validate the JWT
    try:
        validate_jwt(settings, token)
        username = get_username_from_token(token)

        return AuthStatusResponse(
            authenticated=True,
            username=username,
            token=token,
        )
    except HTTPException:
        # JWT validation failed
        return AuthStatusResponse(
            authenticated=False,
            username=None,
            token=token,
        )



from matchbox.common.transform import DisjointSet


class TestDisjointSet:
    def test_disjoint_set_empty(self) -> None:
        dsj = DisjointSet()

        assert dsj.get_components() == []

    def test_disjoint_set_same(self) -> None:
        dsj = DisjointSet()
        dsj.union(1, 1)

        assert dsj.get_components() == [{1}]

    def test_disjoint_set_redundant(self) -> None:
        dsj = DisjointSet()
        dsj.union(1, 2)

        assert dsj.get_components() == [{1, 2}]

        dsj.union(2, 1)

        assert dsj.get_components() == [{1, 2}]

    def test_disjoint_set_union(self) -> None:
        dsj = DisjointSet()
        dsj.union(1, 2)
        dsj.union(3, 4)
        dsj.union(5, 6)

        assert sorted(dsj.get_components()) == [{1, 2}, {3, 4}, {5, 6}]

        dsj.union(2, 3)
        dsj.union(4, 5)

        assert dsj.get_components() == [{1, 2, 3, 4, 5, 6}]

    def test_disjoint_set_add_single(self) -> None:
        dsj = DisjointSet()
        dsj.add(1)

        assert dsj.get_components() == [{1}]

    def test_disjoint_set_add_multiple(self) -> None:
        dsj = DisjointSet()
        dsj.add(1)
        dsj.add(2)
        dsj.add(3)

        # Get components and ensure we have the expected singletons
        components = dsj.get_components()
        assert len(components) == 3
        assert {1} in components
        assert {2} in components
        assert {3} in components

    def test_disjoint_set_add_existing(self) -> None:
        dsj = DisjointSet()
        dsj.add(1)
        dsj.add(1)  # Add the same element again

        assert dsj.get_components() == [{1}]  # Should still be just one component

    def test_disjoint_set_add_after_union(self) -> None:
        dsj = DisjointSet()
        dsj.union(1, 2)
        dsj.add(1)  # Add an element that's already in a union

        assert dsj.get_components() == [{1, 2}]  # Should not change existing structure

    def test_disjoint_set_union_after_add(self) -> None:
        dsj = DisjointSet()
        dsj.add(1)
        dsj.add(2)
        dsj.union(1, 2)

        assert dsj.get_components() == [{1, 2}]

    def test_disjoint_set_mixed_operations(self) -> None:
        dsj = DisjointSet()
        dsj.add(1)
        dsj.add(2)
        dsj.union(3, 4)  # Creates {3,4} implicitly
        dsj.add(5)

        # Check that the expected components are present
        components = dsj.get_components()
        assert len(components) == 4
        assert {1} in components
        assert {2} in components
        assert {3, 4} in components
        assert {5} in components



"""Test deterministic behavior of linkers."""

from collections.abc import Callable
from typing import Any
from unittest.mock import Mock, patch

import polars as pl
import pytest
from splink import SettingsCreator
from splink import blocking_rule_library as brl
from splink import comparison_library as cl
from splink.internals.blocking_rule_creator import BlockingRuleCreator
from splink.internals.comparison_creator import ComparisonCreator

from matchbox.client.models import Model
from matchbox.client.models.linkers.base import Linker
from matchbox.client.models.linkers.deterministic import (
    DeterministicLinker,
    DeterministicSettings,
)
from matchbox.client.models.linkers.splinklinker import SplinkLinker, SplinkSettings
from matchbox.client.models.linkers.weighteddeterministic import (
    WeightedDeterministicLinker,
    WeightedDeterministicSettings,
)
from matchbox.client.queries import Query
from matchbox.client.results import ModelResults
from matchbox.common.factories.entities import FeatureConfig
from matchbox.common.factories.sources import (
    SourceTestkit,
    SourceTestkitParameters,
    linked_sources_factory,
    source_factory,
)

LinkerConfigurator = Callable[[SourceTestkit, SourceTestkit], dict[str, Any]]

# Methodology configuration adapters


def configure_deterministic_linker(
    left_testkit: SourceTestkit, right_testkit: SourceTestkit
) -> dict[str, Any]:
    """Configure settings for DeterministicLinker using only shared fields.

    Args:
        left_testkit: Left SourceTestkit from linked_sources_factory
        right_testkit: Right SourceTestkit from linked_sources_factory

    Returns:
        A dictionary with validated settings for DeterministicLinker
    """
    # Extract field names excluding key and id
    left_fields = {
        c.name
        for c in left_testkit.source_config.index_fields
        if c.name not in ("key", "id")
    }
    right_fields = {
        c.name
        for c in right_testkit.source_config.index_fields
        if c.name not in ("key", "id")
    }
    shared_fields: list[str] = sorted(left_fields & right_fields)

    if not shared_fields:
        raise ValueError("Must have at least one shared field")

    # Build comparison strings
    comparisons: list[str] = []
    for field in shared_fields:
        comparisons.append(f"l.{field} = r.{field}")

    settings_dict = {
        "left_id": "id",
        "right_id": "id",
        "comparisons": comparisons,
    }

    # Validate the settings dictionary
    DeterministicSettings.model_validate(settings_dict)

    return settings_dict


def configure_deterministic_linker_sequential(
    left_testkit: SourceTestkit, right_testkit: SourceTestkit
) -> dict[str, Any]:
    """Configure settings for DeterministicLinker with sequential rounds.

    Args:
        left_testkit: Left SourceTestkit from linked_sources_factory
        right_testkit: Right SourceTestkit from linked_sources_factory

    Returns:
        A dictionary with validated settings for DeterministicLinker using
            sequential rounds
    """
    # Extract field names excluding key and id
    left_fields = {
        c.name
        for c in left_testkit.source_config.index_fields
        if c.name not in ("key", "id")
    }
    right_fields = {
        c.name
        for c in right_testkit.source_config.index_fields
        if c.name not in ("key", "id")
    }
    shared_fields: list[str] = sorted(left_fields & right_fields)

    if not shared_fields:
        raise ValueError("Must have at least one shared field")

    # Build sequential rounds - one comparison per round
    comparisons: list[list[str]] = []
    for field in shared_fields:
        comparisons.append([f"l.{field} = r.{field}"])

    settings_dict = {
        "left_id": "id",
        "right_id": "id",
        "comparisons": comparisons,
    }

    # Meaningfully using rounds
    assert len(comparisons) > 1

    # Validate the settings dictionary
    DeterministicSettings.model_validate(settings_dict)

    return settings_dict


def configure_weighted_deterministic_linker(
    left_testkit: SourceTestkit, right_testkit: SourceTestkit
) -> dict[str, Any]:
    """Configure settings for WeightedDeterministicLinker using only shared fields.

    Args:
        left_testkit: Left source object from linked_sources_factory
        right_testkit: Right source object from linked_sources_factory

    Returns:
        A dictionary with validated settings for WeightedDeterministicLinker
    """
    # Extract field names excluding key and id
    left_fields = {
        c.name
        for c in left_testkit.source_config.index_fields
        if c.name not in ("key", "id")
    }
    right_fields = {
        c.name
        for c in right_testkit.source_config.index_fields
        if c.name not in ("key", "id")
    }
    shared_fields: list[str] = sorted(left_fields & right_fields)

    if not shared_fields:
        raise ValueError("Must have at least one shared field")

    # Build weighted comparisons with equal weights
    weighted_comparisons = []
    for field in shared_fields:
        weighted_comparisons.append(
            {"comparison": f"l.{field} = r.{field}", "weight": 1}
        )

    # Create settings dictionary
    settings_dict = {
        "left_id": "id",
        "right_id": "id",
        "weighted_comparisons": weighted_comparisons,
        "threshold": 1,  # Require all comparisons to match
    }

    # Validate the settings dictionary
    WeightedDeterministicSettings.model_validate(settings_dict)

    return settings_dict


def configure_splink_linker(
    left_testkit: SourceTestkit, right_testkit: SourceTestkit
) -> dict[str, Any]:
    """Configure settings for SplinkLinker using only shared fields.

    Args:
        left_testkit: Left source object from linked_sources_factory
        right_testkit: Right source object from linked_sources_factory

    Returns:
        A dictionary with validated settings for SplinkLinker
    """
    # Extract field names excluding key and id
    left_fields = {
        c.name
        for c in left_testkit.source_config.index_fields
        if c.name not in ("key", "id")
    }
    right_fields = {
        c.name
        for c in right_testkit.source_config.index_fields
        if c.name not in ("key", "id")
    }
    shared_fields: list[str] = sorted(left_fields & right_fields)

    if not shared_fields:
        raise ValueError("Must have at least one shared field")

    deterministic_matching_rules: list[str] = []
    blocking_rules_to_generate_predictions: list[BlockingRuleCreator] = []
    comparisons: list[ComparisonCreator] = []

    for field in shared_fields:
        deterministic_matching_rules.append(f"l.{field} = r.{field}")
        blocking_rules_to_generate_predictions.append(brl.block_on(field))
        comparisons.append(cl.ExactMatch(field).configure(m_probabilities=[1, 0]))

    linker_training_functions = [
        {
            "function": "estimate_probability_two_random_records_match",
            "arguments": {
                "deterministic_matching_rules": deterministic_matching_rules,
                "recall": 1,
            },
        },
        {
            "function": "estimate_u_using_random_sampling",
            "arguments": {"max_pairs": 1e4},
        },
    ]

    # The m parameter is 1 because we're testing in a deterministic system, and
    # many of these tests only have one field, so we can't use expectation
    # maximisation to estimate. For testing raw functionality, fine to use 1
    linker_settings = SettingsCreator(
        link_type="link_only",
        retain_matching_columns=False,
        retain_intermediate_calculation_columns=False,
        blocking_rules_to_generate_predictions=blocking_rules_to_generate_predictions,
        comparisons=comparisons,
    )

    settings_dict = {
        "left_id": "id",
        "right_id": "id",
        "linker_training_functions": linker_training_functions,
        "linker_settings": linker_settings,
        "threshold": None,
    }

    # Validate the settings dictionary
    SplinkSettings.model_validate(settings_dict)

    return settings_dict


LINKERS = [
    pytest.param(
        DeterministicLinker, configure_deterministic_linker, id="Deterministic"
    ),
    pytest.param(
        DeterministicLinker,
        configure_deterministic_linker_sequential,
        id="DeterministicSequential",
    ),
    pytest.param(
        WeightedDeterministicLinker,
        configure_weighted_deterministic_linker,
        id="WeightedDeterministic",
    ),
    pytest.param(SplinkLinker, configure_splink_linker, id="Splink"),
    # Add more linker classes and configuration functions here
]

# Test cases


@pytest.mark.parametrize(("Linker", "configure_linker"), LINKERS)
@patch.object(Query, "data")
def test_exact_match_linking(
    mock_query_run: Mock, Linker: Linker, configure_linker: LinkerConfigurator
) -> None:
    """Test linking with exact matches between sources."""
    # Create sources with the same entities
    features = (
        FeatureConfig(
            name="company",
            base_generator="company",
        ),
        FeatureConfig(
            name="email",
            base_generator="email",
        ),
    )

    configs = (
        SourceTestkitParameters(
            name="source_left",
            features=features,
            n_true_entities=10,
        ),
        SourceTestkitParameters(
            name="source_right",
            features=features,
            n_true_entities=10,  # Same number of entities
        ),
    )

    linked = linked_sources_factory(source_parameters=configs, seed=42)
    left_source = linked.sources["source_left"]
    right_source = linked.sources["source_right"]

    # First left is queried, then right
    mock_query_run.side_effect = [
        pl.from_arrow(left_source.data),
        pl.from_arrow(right_source.data),
    ]

    assert left_source.data.select(["company", "email"]).equals(
        right_source.data.select(["company", "email"])
    )

    # Configure and run the linker
    linker = Model(
        dag=linked.dag,
        name="exact_match_linker",
        description="Linking with exact matches",
        model_class=Linker,
        model_settings=configure_linker(left_source, right_source),
        left_query=Query(left_source.source, dag=linked.dag),
        right_query=Query(right_source.source, dag=linked.dag),
    )
    results: ModelResults = linker.run()

    # Validate results against ground truth
    identical, report = linked.diff_results(
        probabilities=results.probabilities,
        left_clusters=left_source.entities,
        right_clusters=right_source.entities,
        sources=["source_left", "source_right"],
        threshold=0,
    )

    assert identical, f"Expected perfect results but got: {report}"


@pytest.mark.parametrize(("Linker", "configure_linker"), LINKERS)
@patch.object(Query, "data")
def test_exact_match_with_duplicates_linking(
    mock_query_run: Mock, Linker: Linker, configure_linker: LinkerConfigurator
) -> None:
    """Test linking with exact matches between sources, where data is duplicated."""
    # Create sources with the same entities
    features = (
        FeatureConfig(
            name="company",
            base_generator="company",
        ),
        FeatureConfig(
            name="email",
            base_generator="email",
        ),
    )

    configs = (
        SourceTestkitParameters(
            name="source_left",
            features=features,
            n_true_entities=10,
            repetition=1,  # Each entity appears twice
        ),
        SourceTestkitParameters(
            name="source_right",
            features=features,
            n_true_entities=10,  # Same number of entities
            repetition=3,  # Each entity appears four times
        ),
    )

    linked = linked_sources_factory(source_parameters=configs, seed=42)
    left_source = linked.sources["source_left"]
    right_source = linked.sources["source_right"]

    mock_query_run.side_effect = [
        pl.from_arrow(left_source.data),
        pl.from_arrow(right_source.data),
    ]

    # Configure and run the linker
    linker = Model(
        dag=linked.dag,
        name="exact_match_linker",
        description="Linking with exact matches",
        model_class=Linker,
        model_settings=configure_linker(left_source, right_source),
        left_query=Query(left_source, dag=linked.dag),
        right_query=Query(right_source, dag=linked.dag),
    )
    results: ModelResults = linker.run()

    # Validate results against ground truth
    identical, report = linked.diff_results(
        probabilities=results.probabilities,
        left_clusters=left_source.entities,
        right_clusters=right_source.entities,
        sources=["source_left", "source_right"],
        threshold=0,
    )

    assert identical, f"Expected perfect results but got: {report}"


@pytest.mark.parametrize(("Linker", "configure_linker"), LINKERS)
@patch.object(Query, "data")
def test_partial_entity_linking(
    mock_query_run: Mock, Linker: Linker, configure_linker: LinkerConfigurator
) -> None:
    """Test linking when one source contains only a subset of entities.

    This tests that the linker correctly handles when the right source
    only contains a subset of the entities in the left source.
    """
    # Create features for our test sources
    features = (
        FeatureConfig(
            name="company",
            base_generator="company",
        ),
        FeatureConfig(
            name="registration_id",
            base_generator="numerify",
            parameters=(("text", "######"),),
        ),
    )

    # Configure sources - full set on left, half on right
    configs = (
        SourceTestkitParameters(
            name="source_left",
            features=features,
            n_true_entities=10,  # Full set
        ),
        SourceTestkitParameters(
            name="source_right",
            features=features,
            n_true_entities=5,  # Half the entities
        ),
    )

    # Create the linked sources
    linked = linked_sources_factory(source_parameters=configs, seed=42)
    left_source = linked.sources["source_left"]
    right_source = linked.sources["source_right"]

    mock_query_run.side_effect = [
        pl.from_arrow(left_source.data),
        pl.from_arrow(right_source.data),
    ]

    # Configure and run the linker
    linker = Model(
        dag=linked.dag,
        name="partial_match_linker",
        description="Linking with partial entity coverage",
        model_class=Linker,
        model_settings=configure_linker(left_source, right_source),
        left_query=Query(left_source, dag=linked.dag),
        right_query=Query(right_source, dag=linked.dag),
    )
    results = linker.run()

    # Validate results against ground truth
    identical, report = linked.diff_results(
        probabilities=results.probabilities,
        left_clusters=left_source.entities,
        right_clusters=right_source.entities,
        sources=["source_left", "source_right"],
        threshold=0,
    )

    assert identical, f"Expected perfect results but got: {report}"


@pytest.mark.parametrize(("Linker", "configure_linker"), LINKERS)
@patch.object(Query, "data")
def test_no_matching_entities_linking(
    mock_query_run: Mock, Linker: Linker, configure_linker: LinkerConfigurator
) -> None:
    """Test linking when there are no matching entities between sources.

    Verifies linkers behave correctly when there should be no matches.
    """
    # Create two data source with disjoint entities
    features = (
        FeatureConfig(name="company", base_generator="company"),
        FeatureConfig(name="identifier", base_generator="uuid4"),
    )

    configs = (
        SourceTestkitParameters(
            name="source_left",
            features=features,
            n_true_entities=10,
        ),
    )

    linked = linked_sources_factory(source_parameters=configs, seed=314)
    left_source = linked.sources["source_left"]
    right_source = source_factory(
        name="source_right", features=features, n_true_entities=10, seed=159
    )

    mock_query_run.side_effect = [
        pl.from_arrow(left_source.data),
        pl.from_arrow(right_source.data),
    ]

    for column in ("company", "identifier"):
        l_col = set(left_source.data[column].to_pylist())
        r_col = set(right_source.data[column].to_pylist())
        assert l_col.isdisjoint(r_col)

    # Configure and run the linker
    linker = Model(
        dag=linked.dag,
        name="no_match_linker",
        description="Linking with no matching entities",
        model_class=Linker,
        model_settings=configure_linker(left_source, right_source),
        left_query=Query(left_source, dag=linked.dag),
        right_query=Query(right_source, dag=linked.dag),
    )
    results = linker.run()

    # Validate results against ground truth
    identical, report = linked.diff_results(
        probabilities=results.probabilities,
        left_clusters=left_source.entities,
        right_clusters=right_source.entities,
        sources=["source_left", "source_right"],
        threshold=0,
    )

    assert not identical
    assert len(results.probabilities) == 0



"""Worker logic to process user uploads."""

from abc import ABC, abstractmethod
from collections.abc import Generator
from functools import partial
from typing import TYPE_CHECKING, Any

import pyarrow as pa
import redis
from celery import Celery, Task
from celery.exceptions import MaxRetriesExceededError
from celery.utils.log import get_task_logger
from fastapi import UploadFile
from pyarrow import parquet as pq

from matchbox.common.arrow import SCHEMA_INDEX, SCHEMA_RESULTS
from matchbox.common.dtos import (
    ResolutionPath,
    ResolutionType,
)
from matchbox.common.exceptions import MatchboxServerFileError
from matchbox.common.logging import log_mem_usage, logger
from matchbox.server.base import (
    MatchboxBackends,
    MatchboxDBAdapter,
    MatchboxServerSettings,
    get_backend_settings,
    settings_to_backend,
)

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client
else:
    S3Client = Any

celery_logger = get_task_logger(__name__)

# -- Upload trackers --


class UploadTracker(ABC):
    """Upload error tracker."""

    @abstractmethod
    def get(self, upload_id: str) -> str | None:
        """Retrieve error message from tracker."""
        ...

    @abstractmethod
    def set(self, upload_id: str, message: str) -> None:
        """Add error message to tracker."""
        ...


class InMemoryUploadTracker:
    """In-memory error tracker."""

    def __init__(self) -> None:
        """Initialise dictionary."""
        self.tracker: dict[str, str] = {}

    def get(self, upload_id: str) -> str | None:  # noqa: D102
        return self.tracker.get(upload_id)

    def set(self, upload_id: str, message: str) -> None:  # noqa: D102
        self.tracker[upload_id] = message


class RedisUploadTracker:
    """Error tracker backed by Redis."""

    def __init__(
        self, redis_url: str, expiry_minutes: int, key_space: str = "upload"
    ) -> None:
        """Connect Redis and initialise tracker object."""
        self.expiry_minutes = expiry_minutes
        self.redis = redis.Redis.from_url(redis_url)
        self.key_prefix = f"{key_space}:"

    def get(self, upload_id: str) -> str | None:  # noqa: D102
        results = self.redis.get(f"{self.key_prefix}{upload_id}")
        if results:
            return results.decode("utf-8")
        return None

    def set(self, upload_id: str, message: str) -> None:  # noqa: D102
        expiry_seconds = self.expiry_minutes * 60
        self.redis.setex(f"{self.key_prefix}{upload_id}", expiry_seconds, message)


_IN_MEMORY_TRACKER = InMemoryUploadTracker()


def settings_to_upload_tracker(settings: MatchboxServerSettings) -> UploadTracker:
    """Initialise an upload tracker from server settings."""
    match settings.task_runner:
        case "api":
            return _IN_MEMORY_TRACKER
        case "celery":
            return RedisUploadTracker(
                redis_url=settings.redis_uri,
                expiry_minutes=settings.uploads_expiry_minutes,
            )
        case _:
            raise RuntimeError("Unsupported task runner.")


# -- S3 functions --


def file_to_s3(client: S3Client, bucket: str, key: str, file: UploadFile) -> str:
    """Upload a PyArrow Table to S3 and return the key.

    Args:
        client: The S3 client to use.
        bucket: The S3 bucket to upload to.
        key: The key to upload to.
        file: The file to upload.

    Raises:
        MatchboxServerFileError: If the file is not a valid Parquet file or the schema
            does not match the expected schema.

    Returns:
        The key of the uploaded file.
    """
    try:
        file.file.seek(0)
        client.put_object(Bucket=bucket, Key=key, Body=file.file)
    except Exception as e:
        if isinstance(e, MatchboxServerFileError):
            raise
        raise MatchboxServerFileError(message=f"Invalid Parquet file: {str(e)}") from e

    return key


def s3_to_recordbatch(
    client: S3Client, bucket: str, key: str, batch_size: int = 1000
) -> Generator[pa.RecordBatch, None, None]:
    """Download a PyArrow Table from S3 and stream it as RecordBatches."""
    response = client.get_object(Bucket=bucket, Key=key)
    buffer = pa.BufferReader(response["Body"].read())

    parquet_file = pq.ParquetFile(buffer)

    yield from parquet_file.iter_batches(batch_size=batch_size)


# -- Upload tasks --


CELERY_SETTINGS: MatchboxServerSettings = get_backend_settings(
    MatchboxServerSettings().backend_type
)()
CELERY_BACKEND: MatchboxDBAdapter | None = None
CELERY_TRACKER: UploadTracker | None = None

celery = Celery("matchbox", broker=CELERY_SETTINGS.redis_uri)
celery.conf.update(
    # Hard time limit for tasks (in seconds)
    task_time_limit=CELERY_SETTINGS.uploads_expiry_minutes * 60,
    # Only acknowledge task (remove it from queue) after task completion
    task_acks_late=True,
    # Reduce pre-fetching (workers reserving tasks while they're still busy)
    # as it's not ideal for long-running tasks
    prefetch_multiplier=1,
    # Limit number of tasks a worker can complete before restarting
    # This may prevent OOM errors caused by memory not being deallocated properly
    worker_max_tasks_per_child=5,
)


def initialise_celery_worker() -> None:
    """Initialise backend and tracker for celery worker and set up logging."""
    global CELERY_SETTINGS
    global CELERY_BACKEND
    global CELERY_TRACKER

    if not CELERY_BACKEND:
        CELERY_BACKEND = settings_to_backend(CELERY_SETTINGS)
    if not CELERY_TRACKER:
        CELERY_TRACKER = settings_to_upload_tracker(CELERY_SETTINGS)


def process_upload(
    backend: MatchboxDBAdapter,
    tracker: UploadTracker,
    s3_client: S3Client,
    resolution_path: ResolutionPath,
    upload_id: str,
    bucket: str,
    filename: str,
) -> None:
    """Generic task to process uploaded file, usable by FastAPI BackgroundTasks."""
    try:
        resolution = backend.get_resolution(path=resolution_path)

        batches = [
            batch
            for batch in s3_to_recordbatch(
                client=s3_client, bucket=bucket, key=filename
            )
        ]

        # If successful, either backend method marks resolution as complete
        if resolution.resolution_type == ResolutionType.SOURCE:
            backend.insert_source_data(
                path=resolution_path,
                data_hashes=pa.Table.from_batches(batches, schema=SCHEMA_INDEX),
            )
        else:
            backend.insert_model_data(
                path=resolution_path,
                results=pa.Table.from_batches(batches, schema=SCHEMA_RESULTS),
            )

    except Exception as e:
        # After failure, signal to clients they can try again
        backend.unlock_resolution_data(path=resolution_path)

        error_context = {
            "upload_id": upload_id,
            "resolution_path": str(resolution_path),
            "bucket": bucket,
            "key": filename,
        }
        logger.error(
            f"Upload processing failed with context: {error_context}", exc_info=True
        )

        # Attach error to upload ID to inform clients
        tracker.set(upload_id=upload_id, message=str(e))

        raise MatchboxServerFileError(
            message=f"Error: {e}. Context: {error_context}"
        ) from e
    finally:
        try:
            s3_client.delete_object(Bucket=bucket, Key=filename)
        except Exception as delete_error:  # noqa: BLE001
            logger.error(
                f"Failed to delete S3 file {bucket}/{filename}: {delete_error}"
            )


@celery.task(ignore_result=True, bind=True, max_retries=3)
def process_upload_celery(
    self: Task, resolution_path_json: str, upload_id: str, bucket: str, filename: str
) -> None:
    """Celery task to process uploaded file, with only serialisable arguments."""
    initialise_celery_worker()
    resolution_path = ResolutionPath.model_validate_json(resolution_path_json)

    # If using Postgres, we must reset the global database connections
    # to avoid using inherited C-pointers from the parent process (ADBC).
    if CELERY_SETTINGS.backend_type == MatchboxBackends.POSTGRES:
        from matchbox.server.postgresql.db import MBDB  # noqa: PLC0415

        MBDB._disconnect_adbc()

    celery_logger.info(
        "Uploading data for resolution %s, ID %s", str(resolution_path), upload_id
    )
    log_mem_usage()

    upload_function = partial(
        process_upload,
        backend=CELERY_BACKEND,
        tracker=CELERY_TRACKER,
        s3_client=CELERY_BACKEND.settings.datastore.get_client(),
    )

    try:
        upload_function(
            resolution_path=resolution_path,
            upload_id=upload_id,
            bucket=bucket,
            filename=filename,
        )
        celery_logger.info(
            "Upload complete for %s, ID %s", str(resolution_path), upload_id
        )

    except Exception as exc:  # noqa: BLE001
        celery_logger.error(
            "Upload failed for resolution %s, ID %s. Retrying...",
            str(resolution_path),
            upload_id,
        )
        try:
            raise self.retry(exc=exc) from None
        except MaxRetriesExceededError:
            CELERY_TRACKER.set(upload_id, f"Max retries exceeded: {exc}")
            raise

    finally:
        log_mem_usage()

        # Cleanup connections
        if CELERY_SETTINGS.backend_type == MatchboxBackends.POSTGRES:
            from matchbox.server.postgresql.db import MBDB  # noqa: PLC0415

            MBDB._disconnect()
            MBDB._disconnect_adbc()



"""Functions and classes to define, run and register models."""

import inspect
import json
from collections.abc import Callable
from functools import wraps
from typing import TYPE_CHECKING, Any, ParamSpec, TypeVar, overload

from polars import DataFrame

from matchbox.client import _handler
from matchbox.client.models import dedupers, linkers
from matchbox.client.models.dedupers.base import Deduper, DeduperSettings
from matchbox.client.models.linkers.base import Linker, LinkerSettings
from matchbox.client.queries import Query
from matchbox.client.results import ModelResults
from matchbox.common.dtos import (
    ModelConfig,
    ModelResolutionName,
    ModelResolutionPath,
    ModelType,
    Resolution,
    ResolutionType,
    SourceResolutionName,
)
from matchbox.common.exceptions import MatchboxResolutionNotFoundError
from matchbox.common.hash import hash_model_results
from matchbox.common.logging import logger, profile_time
from matchbox.common.transform import threshold_float_to_int, threshold_int_to_float

if TYPE_CHECKING:
    from matchbox.client.dags import DAG
    from matchbox.client.sources import Source
else:
    DAG = Any
    Source = Any

P = ParamSpec("P")
R = TypeVar("R")
T = TypeVar("T")


_MODEL_CLASSES = {
    **{name: obj for name, obj in inspect.getmembers(dedupers, inspect.isclass)},
    **{name: obj for name, obj in inspect.getmembers(linkers, inspect.isclass)},
}


def add_model_class(ModelClass: type[Linker] | type[Deduper]) -> None:
    """Add custom deduper or linker."""
    if issubclass(ModelClass, Linker) or issubclass(ModelClass, Deduper):
        _MODEL_CLASSES[ModelClass.__name__] = ModelClass
    else:
        raise ValueError("The argument is not a proper subclass of Deduper or Linker.")


def post_run(method: Callable[..., T]) -> Callable[..., T]:
    """Decorator to ensure that a method is called after model run.

    Raises:
        RuntimeError: If run hasn't happened.
    """

    @wraps(method)
    def wrapper(self: "Model", *args: Any, **kwargs: Any) -> T:
        if self.results is None:
            raise RuntimeError(
                "The model must be run before attempting this operation."
            )
        return method(self, *args, **kwargs)

    return wrapper


class Model:
    """Unified model class for both linking and deduping operations."""

    @overload
    def __init__(
        self,
        name: str,
        dag: DAG,
        model_class: type[Deduper],
        model_settings: DeduperSettings | dict,
        left_query: Query,
        right_query: None = None,
        truth: float = 1.0,
        description: str | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        dag: DAG,
        name: str,
        model_class: type[Linker],
        model_settings: LinkerSettings | dict,
        left_query: Query,
        right_query: Query,
        truth: float = 1.0,
        description: str | None = None,
    ) -> None: ...

    def __init__(
        self,
        dag: DAG,
        name: str,
        model_class: type[Deduper] | type[Linker] | str,
        model_settings: DeduperSettings | LinkerSettings | dict,
        left_query: Query,
        right_query: Query | None = None,
        truth: float = 1.0,
        description: str | None = None,
    ):
        """Create a new model instance.

        Args:
            dag: DAG containing this model.
            name: Unique name for the model
            truth: Truth threshold. Defaults to 1.0. Can be set later after analysis.
            model_class: Class of Linker or Deduper, or its name.
            model_settings: Appropriate settings object to pass to model class.
            left_query: The query that will get the data to deduplicate, or the data to
                link on the left.
            right_query: The query that will get the data to link on the right.
            description: Optional description of the model
        """
        self.dag = dag
        self.name = name
        self.description = description
        self._truth: int = threshold_float_to_int(truth)
        self.left_query = left_query
        self.right_query = right_query
        self.results: ModelResults | None = None

        if isinstance(model_class, str):
            self.model_class: type[Linker | Deduper] = _MODEL_CLASSES[model_class]
        else:
            self.model_class = model_class
        self.model_instance = self.model_class(settings=model_settings)

        self.model_type: ModelType = (
            ModelType.LINKER
            if issubclass(self.model_class, Linker)
            else ModelType.DEDUPER
        )

        if isinstance(model_settings, dict):
            SettingsClass = self.model_instance.__annotations__["settings"]
            self.model_settings = SettingsClass(**model_settings)
        else:
            self.model_settings = model_settings

    @property
    def config(self) -> ModelConfig:
        """Generate config DTO from Model."""
        return ModelConfig(
            type=self.model_type,
            model_class=self.model_class.__name__,
            model_settings=self.model_settings.model_dump_json(),
            left_query=self.left_query.config,
            right_query=self.right_query.config if self.right_query else None,
        )

    @property
    def sources(self) -> set[SourceResolutionName]:
        """Set of source names upstream of this node."""
        left_input = self.dag.nodes[self.left_query.config.point_of_truth]
        model_sources = left_input.sources
        if self.right_query:
            right_input = self.dag.nodes[self.right_query.config.point_of_truth]
            model_sources.update(right_input.sources)

        return model_sources

    @post_run
    def to_resolution(self) -> Resolution:
        """Convert to Resolution for API calls."""
        return Resolution(
            description=self.description,
            truth=self._truth,
            resolution_type=ResolutionType.MODEL,
            config=self.config,
            fingerprint=hash_model_results(self.results.probabilities),
        )

    @classmethod
    def from_resolution(
        cls,
        resolution: Resolution,
        resolution_name: str,
        dag: DAG,
    ) -> "Model":
        """Reconstruct from Resolution."""
        if resolution.resolution_type != ResolutionType.MODEL:
            raise ValueError("Resolution must be of type 'model'")

        return cls(
            dag=dag,
            name=ModelResolutionName(resolution_name),
            description=resolution.description,
            model_class=resolution.config.model_class,
            model_settings=json.loads(resolution.config.model_settings),
            left_query=Query.from_config(resolution.config.left_query, dag=dag),
            right_query=Query.from_config(resolution.config.right_query, dag=dag)
            if resolution.config.right_query
            else None,
            truth=threshold_int_to_float(resolution.truth),
        )

    @property
    def resolution_path(self) -> ModelResolutionPath:
        """Returns the model resolution path."""
        return ModelResolutionPath(
            collection=self.dag.name,
            run=self.dag.run,
            name=self.name,
        )

    @property
    def truth(self) -> float | None:
        """Returns the truth threshold for the model as a float."""
        if self._truth is not None:
            return threshold_int_to_float(self._truth)
        return None

    @truth.setter
    def truth(self, truth: float) -> None:
        """Set the truth threshold for the model."""
        self._truth = threshold_float_to_int(truth)

    def delete(self, certain: bool = False) -> bool:
        """Delete the model from the database."""
        logger.info(f"Deleting {self.name}")
        result = _handler.delete_resolution(path=self.resolution_path, certain=certain)
        return result.success

    @profile_time(attr="name")
    def compute_probabilities(
        self, left_df: DataFrame, right_df: DataFrame | None = None
    ) -> DataFrame:
        """Run model instance against data."""
        if self.config.type == ModelType.LINKER:
            self.model_instance.prepare(left_df, right_df)
            probabilities = self.model_instance.link(left=left_df, right=right_df)
        else:
            self.model_instance.prepare(left_df)
            probabilities = self.model_instance.dedupe(data=left_df)

        return probabilities

    def run(
        self,
        left_data: DataFrame | None = None,
        right_data: DataFrame | None = None,
        for_validation: bool = False,
    ) -> ModelResults:
        """Execute the model pipeline and return results.

        Args:
            left_data (optional): Pre-fetched query data to deduplicate if the model is
                a deduper, or link on the left if the model is a linker.
            right_data (optional): Pre-fetched query data to link on the right, if the
                model is a linker. If the model is a deduper, this argument is ignored.
            for_validation: Whether to download and store extra data to explore and
                score results.
        """
        log_prefix = f"Run {self.name}"
        logger.info("Executing left query", prefix=log_prefix)

        left_df = (
            left_data
            if left_data is not None
            else self.left_query.data(return_leaf_id=for_validation)
        )
        right_df = None

        if self.config.type == ModelType.LINKER:
            logger.info("Executing right query", prefix=log_prefix)
            right_df = (
                right_data
                if right_data is not None
                else self.right_query.data(return_leaf_id=for_validation)
            )

        logger.info("Running model logic", prefix=log_prefix)
        probabilities = self.compute_probabilities(left_df, right_df)

        if for_validation:
            self.results = ModelResults(
                probabilities=probabilities,
                left_root_leaf=self.left_query.leaf_id,
                right_root_leaf=self.right_query.leaf_id
                if right_df is not None
                else None,
            )
        else:
            self.results = ModelResults(probabilities=probabilities)

        return self.results

    @post_run
    @profile_time(attr="name")
    def sync(self) -> None:
        """Send the model config and results to the server.

        Not resistant to race conditions: only one client should call sync at a time.
        """
        log_prefix = f"Sync {self.name}"
        resolution = self.to_resolution()
        try:
            existing_resolution = _handler.get_resolution(path=self.resolution_path)
            logger.info("Found existing resolution", prefix=log_prefix)
        except MatchboxResolutionNotFoundError:
            existing_resolution = None

        if existing_resolution:
            if (existing_resolution.fingerprint == resolution.fingerprint) and (
                existing_resolution.config.parents == resolution.config.parents
            ):
                logger.info("Updating existing resolution", prefix=log_prefix)
                # Assumes that resolution hasn't been deleted or made incompatible
                # Else, server will error
                _handler.update_resolution(
                    resolution=resolution, path=self.resolution_path
                )
            else:
                logger.info(
                    "Update not possible. Deleting existing resolution",
                    prefix=log_prefix,
                )
                # Assumes that resolution hasn't been deleted, else server will error
                _handler.delete_resolution(path=self.resolution_path, certain=True)
                existing_resolution = None

        if not existing_resolution:
            logger.info("Creating new resolution", prefix=log_prefix)
            # Assumes that resolution hasn't since been re-created.
            # Else, server will error
            _handler.create_resolution(resolution=resolution, path=self.resolution_path)
            logger.info("Setting data for new resolution", prefix=log_prefix)
            # Assumes resolution has not been deleted or made incompatible
            _handler.set_data(
                path=self.resolution_path, data=self.results.probabilities
            )

    def download_results(self) -> ModelResults:
        """Retrieve results associated with the model from the database."""
        results = _handler.get_results(name=self.name)
        return ModelResults(probabilities=results)

    def query(self, *sources: Source, **kwargs: Any) -> Query:
        """Generate a query for this model."""
        return Query(*sources, **kwargs, model=self, dag=self.dag)

    def clear_data(self) -> None:
        """Deletes data computed for node."""
        self.results = None



import pyarrow as pa
import pytest
from sqlalchemy import func, select
from sqlalchemy.dialects.postgresql import BIGINT, TEXT

from matchbox.server.postgresql import MatchboxPostgres
from matchbox.server.postgresql.db import MBDB
from matchbox.server.postgresql.utils.db import ingest_to_temporary_table


@pytest.mark.docker
def test_ingest_to_temporary_table(
    matchbox_postgres: MatchboxPostgres,  # will drop dummy table
) -> None:
    """Test temporary table creation, data ingestion, and automatic cleanup."""
    # Create sample arrow data
    data = pa.Table.from_pylist(
        [
            {"id": 1, "value": "test1"},
            {"id": 2, "value": "test2"},
        ]
    )

    schema_name = MBDB.MatchboxBase.metadata.schema
    table_name = "test_temp_ingest"

    # Define the column types for the temporary table
    column_types = {
        "id": BIGINT(),
        "value": TEXT(),
    }

    # Use the context manager to create and populate a temporary table
    with (
        ingest_to_temporary_table(
            table_name=table_name,
            schema_name=schema_name,
            data=data,
            column_types=column_types,
        ) as temp_table,
        MBDB.get_session() as session,
    ):
        # Verify the table exists and has the expected data
        # Check that the table exists using SQLAlchemy syntax
        result = session.execute(select(func.count()).select_from(temp_table)).scalar()
        assert result == 2

        # Check a specific value using SQLAlchemy syntax
        value = session.execute(
            select(temp_table.c.value).where(temp_table.c.id == 1)
        ).scalar()
        assert value == "test1"

    # After context exit, verify the table no longer exists
    with (
        MBDB.get_session() as session,
        pytest.raises(Exception),  # Should fail as table is dropped # noqa: B017
    ):
        session.execute(select(func.count()).select_from(temp_table))



"""Create schema without tables.

Revision ID: 40a8e5ed48f2
Revises: This is the first migration
"""

from collections.abc import Sequence

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "40a8e5ed48f2"
# down_revision=None indicates that this is the first migration file
down_revision: str | None = None
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    op.execute("CREATE SCHEMA mb")
    op.execute('CREATE EXTENSION IF NOT EXISTS "uuid-ossp"')
    op.execute('CREATE EXTENSION IF NOT EXISTS "pgcrypto"')


def downgrade() -> None:
    """Downgrade schema."""
    op.execute("DROP SCHEMA IF EXISTS mb CASCADE")
    op.execute('DROP EXTENSION IF EXISTS "uuid-ossp"')
    op.execute('DROP EXTENSION IF EXISTS "pgcrypto"')



"""Server management commands for Matchbox CLI."""

import typer
from rich import print

from matchbox.client import _handler

app = typer.Typer(help="Manage Matchbox server")


@app.command()
def health() -> None:
    """Checks the health of the Matchbox server."""
    response = _handler.healthcheck()
    print(response)


@app.command()
def delete_orphans() -> None:
    """Deletes orphans from Matchbox database.

    Orphan clusters are clusters that are not linked to any other table, because they
    have become isolated as a result of the change or removal of resolutions.
    This command will remove them from the database.
    """
    response = _handler.delete_orphans()
    print(response)



"""Authentication management commands for Matchbox CLI."""

import typer
from rich import print

from matchbox.client import _handler

app = typer.Typer(help="Manage Matchbox authentication")


@app.command()
def status() -> None:
    """Checks the authentication status."""
    response = _handler.auth_status()
    print(response)



"""Migrate to SQL-first insertion logic.

Revision ID: 3dfe29cecf2f
Revises: 58ee4e19fb7d
Create Date: 2026-01-12 10:08:32.246473

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "3dfe29cecf2f"
down_revision: str | None = "58ee4e19fb7d"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    connection = op.get_bind()
    metadata = sa.MetaData(schema="mb")
    pk_space = sa.Table("pk_space", metadata, autoload_with=connection)

    # Migrate next IDs from PKSpace to sequences
    next_ids = connection.execute(
        sa.select(pk_space.c.next_cluster_id, pk_space.c.next_cluster_keys_id)
    ).one_or_none()

    if next_ids:
        connection.execute(
            sa.select(
                sa.func.setval(
                    sa.func.pg_get_serial_sequence("mb.clusters", "cluster_id"),
                    next_ids.next_cluster_id,
                    False,
                ),
            )
        )

        connection.execute(
            sa.select(
                sa.func.setval(
                    sa.func.pg_get_serial_sequence("mb.cluster_keys", "key_id"),
                    next_ids.next_cluster_keys_id,
                    False,
                ),
            )
        )

    # Drop PKSpace
    op.drop_table("pk_space", schema="mb")

    # Create constraint on Contains
    op.create_unique_constraint(
        "contains_unique_root_leaf", "contains", ["root", "leaf"], schema="mb"
    )


def downgrade() -> None:
    """Downgrade schema."""
    connection = op.get_bind()
    metadata = sa.MetaData(schema="mb")

    # Drop constraint on Contains
    op.drop_constraint(
        "contains_unique_root_leaf", "contains", schema="mb", type_="unique"
    )

    # Create PKSpace
    op.create_table(
        "pk_space",
        sa.Column("id", sa.BIGINT(), autoincrement=True, nullable=False),
        sa.Column("next_cluster_id", sa.BIGINT(), autoincrement=False, nullable=False),
        sa.Column(
            "next_cluster_keys_id", sa.BIGINT(), autoincrement=False, nullable=False
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_space_pkey")),
        schema="mb",
    )

    # Migrate next IDs from sequences to PKSpace
    pk_space = sa.Table("pk_space", metadata, autoload_with=connection)
    next_cluster_id = connection.execute(
        sa.select(
            sa.func.nextval(sa.func.pg_get_serial_sequence("mb.clusters", "cluster_id"))
        )
    ).scalar()
    next_cluster_key_id = connection.execute(
        sa.select(
            sa.func.nextval(sa.func.pg_get_serial_sequence("mb.cluster_keys", "key_id"))
        )
    ).scalar()

    connection.execute(
        sa.insert(pk_space).values(
            id=1,
            next_cluster_id=next_cluster_id,
            next_cluster_keys_id=next_cluster_key_id,
        )
    )



"""Main application for entity resolution evaluation."""

import logging
from collections import deque
from pathlib import Path

from pydantic import BaseModel, ConfigDict, Field
from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical
from textual.reactive import reactive
from textual.timer import Timer
from textual.widgets import Footer, Label

from matchbox.client import _handler
from matchbox.client._settings import settings
from matchbox.client.cli.eval.modals import HelpModal, NoSamplesModal
from matchbox.client.cli.eval.widgets.assignment import AssignmentBar
from matchbox.client.cli.eval.widgets.styling import get_group_style
from matchbox.client.cli.eval.widgets.table import ComparisonDisplayTable
from matchbox.client.dags import DAG
from matchbox.client.eval import EvaluationItem, create_judgement, get_samples
from matchbox.common.dtos import ModelResolutionPath
from matchbox.common.exceptions import MatchboxClientSettingsException

logger = logging.getLogger(__name__)


class CLIEvaluationSession(BaseModel):
    """CLI evaluation session state.

    Used by queue to store items with their assignments.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    item: EvaluationItem
    assignments: dict[int, str] = Field(default_factory=dict)


class EvaluationQueue:
    """Deque-based queue with current session always at front."""

    def __init__(self) -> None:
        """Initialise the queue."""
        self.sessions: deque[CLIEvaluationSession] = deque()

    @property
    def current(self) -> CLIEvaluationSession | None:
        """Get the current session (always at index 0)."""
        return self.sessions[0] if self.sessions else None

    @property
    def total_count(self) -> int:
        """Total number of sessions in queue."""
        return len(self.sessions)

    def skip_current(self) -> None:
        """Move current to back of queue."""
        if len(self.sessions) > 1:
            self.sessions.rotate(-1)

    def remove_current(self) -> CLIEvaluationSession | None:
        """Remove and return current session."""
        return self.sessions.popleft() if self.sessions else None

    def add_sessions(self, sessions: list[CLIEvaluationSession]) -> int:
        """Add new sessions to queue, preventing duplicates.

        Returns:
            Number of unique sessions added.
        """
        if not sessions:
            return 0

        existing_ids = {tuple(sorted(session.item.leaves)) for session in self.sessions}
        unique_items = [
            session
            for session in sessions
            if tuple(sorted(session.item.leaves)) not in existing_ids
        ]

        if unique_items:
            self.sessions.extend(unique_items)

        return len(unique_items)


class EntityResolutionApp(App):
    """Main Textual application for entity resolution evaluation."""

    CSS_PATH = Path(__file__).parent / "styles.tcss"
    TITLE = "Matchbox evaluate"
    SUB_TITLE = "match labelling tool"

    # Triggered by action_* methods
    BINDINGS = [
        ("shift+right", "skip", "Skip"),
        ("space", "submit", "Submit"),
        ("escape", "clear", "Clear"),
        ("question_mark,f1", "show_help", "Help"),
        ("ctrl+q,ctrl+c", "quit", "Quit"),
    ]

    # Reactive variables that trigger UI updates
    status: reactive[tuple[str, str]] = reactive(("○ Ready", "dim"))
    current_item: reactive[EvaluationItem | None] = reactive(None)
    current_assignments: reactive[dict[int, str]] = reactive({}, init=False)

    sample_limit: int
    resolution: ModelResolutionPath
    user_name: str
    dag: DAG
    session_tag: str | None
    sample_file: str | None
    show_help: bool

    queue: EvaluationQueue
    timer: Timer | None = None

    def __init__(
        self,
        user: str,
        num_samples: int = 5,
        resolution: ModelResolutionPath | None = None,
        dag: DAG | None = None,
        session_tag: str | None = None,
        sample_file: str | None = None,
        show_help: bool = False,
        scroll_debounce_delay: float | None = 0.3,
    ) -> None:
        """Initialise the entity resolution app.

        Args:
            resolution: The model resolution to evaluate
            num_samples: Number of clusters to sample for evaluation
            user: Username for authentication (overrides settings)
            dag: Pre-loaded DAG with warehouse location attached
            session_tag: String to use for tagging judgements
            sample_file: Path to pre-compiled sample file.
                If set, ignores resolutions.
            show_help: Whether to show help on start
            scroll_debounce_delay: Delay before updating column headers after scroll.
                Set to None to disable debouncing (useful for tests).
        """
        super().__init__()

        if not (resolution or sample_file):
            raise ValueError("Either a resolution or a sample file must be set")

        self.resolution = resolution
        self.user_name = user
        self.sample_limit = num_samples
        self.dag = dag
        self.session_tag = session_tag
        self.sample_file = sample_file

        self.show_help = show_help
        self._scroll_debounce_delay = scroll_debounce_delay

        self.queue = EvaluationQueue()

    # Lifecycle methods
    def compose(self) -> ComposeResult:
        """Compose the main application UI."""
        yield Vertical(
            Horizontal(
                Label("-", id="current-group-label"),
                AssignmentBar(id="assignment-bar"),
                Label(id="status-right"),
                id="status-bar",
                classes="status-bar",
            ),
            ComparisonDisplayTable(
                id="record-table", scroll_debounce_delay=self._scroll_debounce_delay
            ),
            id="main-container",
        )
        yield Footer()

    async def on_mount(self) -> None:
        """Initialise the application."""
        await self.authenticate()

        if self.dag is None:
            raise RuntimeError(
                "DAG not loaded. EntityResolutionApp requires a pre-loaded DAG."
            )

        await self.load_samples()

        if self.queue.total_count == 0:
            await self._handle_no_samples()
            return

        self._load_current_item()

        if self.show_help:
            self.call_after_refresh(self.action_show_help)

    # Message handlers
    def on_comparison_display_table_assignment_made(
        self, message: ComparisonDisplayTable.AssignmentMade
    ) -> None:
        """Update reactive assignments when table reports an assignment."""
        # Create new dict to trigger reactivity
        new_assignments = {
            **self.current_assignments,
            message.column_idx: message.group,
        }
        self.current_assignments = new_assignments

        # Save to queue item
        if current := self.queue.current:
            current.assignments = new_assignments

        # Update assignment bar
        _, colour = get_group_style(message.group)
        assignment_bar = self.query_one("#assignment-bar", AssignmentBar)
        assignment_bar.set_position(message.column_idx, message.group.upper(), colour)

    def on_comparison_display_table_current_group_changed(
        self, message: ComparisonDisplayTable.CurrentGroupChanged
    ) -> None:
        """Update current group label when table's current group changes."""
        # Update current group label
        if message.group:
            _, colour = get_group_style(message.group)
            label_text = f"[{colour}]{message.group.upper()}[/{colour}]"
        else:
            label_text = "-"
        self.query_one("#current-group-label", Label).update(label_text)

    # Reactive watchers
    def watch_status(self) -> None:
        """React to status changes."""
        self._update_status_labels()

    def watch_current_item(self, item: EvaluationItem | None) -> None:
        """React to item changes - propagate to table and reset assignment bar."""
        table = self.query_one(ComparisonDisplayTable)
        table.current_item = item

        # Reset assignment bar for new item
        if item:
            unique_record_groups = item.get_unique_record_groups()
            num_columns = len(unique_record_groups)
            assignment_bar = self.query_one("#assignment-bar", AssignmentBar)
            assignment_bar.reset(num_columns)

            # Load existing assignments into the bar
            if self.current_assignments:
                for col_idx, group in self.current_assignments.items():
                    _, colour = get_group_style(group)
                    assignment_bar.set_position(col_idx, group.upper(), colour)

        # Reset current group label
        self.query_one("#current-group-label", Label).update("-")

    def watch_current_assignments(self, assignments: dict[int, str]) -> None:
        """React to assignment changes - propagate to table."""
        table = self.query_one(ComparisonDisplayTable)
        table.current_assignments = assignments

    # Private methods
    def _load_current_item(self) -> None:
        """Load current queue item into the display."""
        current = self.queue.current
        if current:
            # Set reactive properties (will propagate to table via watchers)
            self.current_assignments = current.assignments.copy()
            self.current_item = current.item

    def _update_status_labels(self) -> None:
        """Update right status label."""
        self.query_one("#status-right", Label).update(self._build_status_right())

    def _build_status_right(self) -> str:
        """Build right status text with status indicator."""
        message, colour = self.status
        if message:
            return f"[{colour}]{message}[/]"
        return "[dim]○ Ready[/]"

    async def _handle_no_samples(self) -> None:
        """Handle empty queue state."""
        self._update_status("◯ No data", "yellow")
        logger.warning(f"No samples available for resolution '{self.resolution}'.")
        await self.action_show_no_samples()

    def _update_status(
        self,
        message: str,
        colour: str = "dim",
        clear_after: float | None = None,
    ) -> None:
        """Update status message with optional auto-clear.

        Args:
            message: Status message to display
            colour: Colour for the message
            clear_after: Seconds after which to auto-clear
        """
        if self.timer:
            self.timer.stop()

        self.status = (message, colour)

        if clear_after:
            self.timer = self.set_timer(
                clear_after, lambda: self._update_status("○ Ready", "dim")
            )

    # Public methods
    async def authenticate(self) -> None:
        """Authenticate with the server."""
        user_name = self.user_name or settings.user
        if not user_name:
            raise MatchboxClientSettingsException("User name is unset.")

        self.user_name = _handler.login(user_name=user_name)

    async def load_samples(self) -> None:
        """Load evaluation samples from the server."""
        needed = max(0, self.sample_limit - self.queue.total_count)

        if needed <= 0:
            return

        self._update_status("⚡ Fetching", "yellow")
        logger.info(
            "Fetching %s samples to reach limit of %s",
            needed,
            self.sample_limit,
        )

        new_samples_dict = None
        try:
            new_samples_dict = get_samples(
                n=needed,
                resolution=self.resolution.name if self.resolution else None,
                user_name=self.user_name,
                dag=self.dag,
                sample_file=self.sample_file,
            )
        except Exception as e:  # noqa: BLE001
            logger.warning(f"Failed to fetch samples: {type(e).__name__}: {e}")

        if new_samples_dict:
            # Wrap evaluation items in CLI sessions
            sessions = [
                CLIEvaluationSession(item=item) for item in new_samples_dict.values()
            ]
            self.queue.add_sessions(sessions)

        if self.queue.total_count == 0:
            await self._handle_no_samples()
        else:
            self._update_status("✓ Ready", "green")

    # Action methods (public interface)
    async def action_skip(self) -> None:
        """Skip current entity (moves to back of queue)."""
        if self.queue.total_count > 1:
            self.queue.skip_current()
            self._load_current_item()
        else:
            await self.action_submit()
            if self.queue.total_count == 0:
                self.exit()

    async def action_submit(self) -> None:
        """Submit current entity if fully painted."""
        current = self.queue.current

        if not current:
            return

        # Check if all columns are assigned
        unique_record_groups = current.item.get_unique_record_groups()
        is_painted = len(current.assignments) == len(unique_record_groups)
        if not is_painted:
            self._update_status("⚠ Incomplete", "yellow", clear_after=3.0)
            return

        try:
            if self.user_name is None:
                raise RuntimeError("User name is not set")
            judgement = create_judgement(
                item=current.item,
                assignments=current.assignments,
                user_name=self.user_name,
                tag=self.session_tag,
            )
            _handler.send_eval_judgement(judgement)
        except Exception as exc:
            self._update_status("⚠ Send failed", "red", clear_after=4.0)
            logger.exception(f"Failed to submit: {exc}")
            return

        self.queue.remove_current()
        await self.load_samples()
        self._load_current_item()
        self._update_status("✓ Sent", "green", clear_after=2.0)

    async def action_clear(self) -> None:
        """Clear current entity's group assignments."""
        if self.queue.current:
            # Clear reactive assignments (will propagate to table)
            self.current_assignments = {}

            # Update queue item
            self.queue.current.assignments = {}

            # Reset assignment bar and current group label
            if self.current_item:
                unique_record_groups = self.current_item.get_unique_record_groups()
                assignment_bar = self.query_one("#assignment-bar", AssignmentBar)
                assignment_bar.reset(len(unique_record_groups))
            self.query_one("#current-group-label", Label).update("-")

    async def action_show_help(self) -> None:
        """Show the help modal."""
        self.push_screen(HelpModal())

    async def action_show_no_samples(self) -> None:
        """Show the no samples modal."""
        self.push_screen(NoSamplesModal())

    async def action_quit(self) -> None:
        """Quit the application."""
        self.exit()



"""Add PKSpace table.

Revision ID: beba75a24962
Revises: f3c9279437f4
Create Date: 2025-05-01 16:53:16.071565

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "beba75a24962"
down_revision: str | None = "f3c9279437f4"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    op.create_table(
        "pk_space",
        sa.Column("id", sa.BIGINT(), nullable=False),
        sa.Column("next_cluster_id", sa.BIGINT(), nullable=False),
        sa.Column("next_cluster_source_pk_id", sa.BIGINT(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        schema="mb",
    )

    op.alter_column(
        "sources",
        "resolution_id",
        existing_type=sa.BIGINT(),
        nullable=False,
        schema="mb",
    )


def downgrade() -> None:
    """Downgrade schema."""
    op.drop_table("pk_space", schema="mb")
    op.alter_column(
        "sources",
        "resolution_id",
        existing_type=sa.BIGINT(),
        nullable=True,
        schema="mb",
    )



"""All client-side functionalities of Matchbox."""

from matchbox.client.dags import DAG
from matchbox.client.locations import RelationalDBLocation

__all__ = (
    "DAG",
    "RelationalDBLocation",
)



"""Minimal test coverage for DataTypes with Array and Struct support."""

import json

import polars as pl
import pytest

from matchbox.common.datatypes import DataTypes


class TestSimpleTypes:
    """Test creation and serialisation of simple types."""

    @pytest.mark.parametrize(
        ("dtype", "expected_value"),
        [
            (DataTypes.STRING, "String"),
            (DataTypes.INT64, "Int64"),
            (DataTypes.FLOAT64, "Float64"),
            (DataTypes.BOOLEAN, "Boolean"),
            (DataTypes.DATETIME, "Datetime"),
        ],
    )
    def test_simple_type_serialisation(
        self, dtype: DataTypes, expected_value: str
    ) -> None:
        """Simple types serialise to their string name."""
        assert dtype.value == expected_value

    @pytest.mark.parametrize(
        "dtype",
        [DataTypes.STRING, DataTypes.INT64, DataTypes.FLOAT64],
    )
    def test_simple_type_roundtrip(self, dtype: DataTypes) -> None:
        """Simple types can round-trip through their string representation."""
        deserialised = DataTypes.model_validate(dtype.value)
        assert deserialised.value == dtype.value


class TestListTypes:
    """Test List container types."""

    @pytest.mark.parametrize(
        ("inner_type", "expected_json"),
        [
            (DataTypes.STRING, '{"type": "List", "inner": "String"}'),
            (DataTypes.INT64, '{"type": "List", "inner": "Int64"}'),
        ],
    )
    def test_list_serialisation(
        self, inner_type: DataTypes, expected_json: str
    ) -> None:
        """Lists serialise with their inner type."""
        list_type = DataTypes.LIST(inner_type)
        assert list_type.value == expected_json

    def test_list_roundtrip(self) -> None:
        """List types can round-trip through JSON."""
        original = DataTypes.LIST(DataTypes.FLOAT64)
        deserialised = DataTypes.model_validate_json(original.value)
        assert deserialised.value == original.value

    def test_nested_list(self) -> None:
        """Nested lists serialise correctly."""
        nested = DataTypes.LIST(DataTypes.LIST(DataTypes.STRING))
        expected = '{"type": "List", "inner": {"type": "List", "inner": "String"}}'
        assert nested.value == expected


class TestArrayTypes:
    """Test Array container types with fixed shape."""

    @pytest.mark.parametrize(
        ("inner_type", "shape", "expected_json"),
        [
            (DataTypes.INT64, 3, '{"type": "Array", "inner": "Int64", "shape": [3]}'),
            (
                DataTypes.FLOAT32,
                5,
                '{"type": "Array", "inner": "Float32", "shape": [5]}',
            ),
            (
                DataTypes.INT64,
                (3,),
                '{"type": "Array", "inner": "Int64", "shape": [3]}',
            ),
        ],
    )
    def test_array_serialisation(
        self, inner_type: DataTypes, shape: int | tuple[int, ...], expected_json: str
    ) -> None:
        """Arrays serialise with inner type and shape."""
        array_type = DataTypes.ARRAY(inner_type, shape=shape)
        assert array_type.value == expected_json

    def test_array_roundtrip(self) -> None:
        """Array types can round-trip through JSON."""
        original = DataTypes.ARRAY(DataTypes.INT32, shape=10)
        deserialised = DataTypes.model_validate_json(original.value)
        assert deserialised.value == original.value
        assert deserialised.shape == (10,)

    def test_array_of_lists(self) -> None:
        """Arrays can contain Lists."""
        array_of_lists = DataTypes.ARRAY(DataTypes.LIST(DataTypes.STRING), shape=2)
        deserialised = DataTypes.model_validate_json(array_of_lists.value)
        assert deserialised.shape == (2,)
        assert deserialised.inner is not None
        assert deserialised.inner.base_type.value == "List"


class TestStructTypes:
    """Test Struct types with named fields."""

    def test_simple_struct_serialisation(self) -> None:
        """Structs serialise with their fields."""
        struct = DataTypes.STRUCT(
            fields={"name": DataTypes.STRING, "age": DataTypes.INT64}
        )
        parsed = json.loads(struct.value)
        assert parsed["type"] == "Struct"
        assert parsed["fields"]["name"] == "String"
        assert parsed["fields"]["age"] == "Int64"

    def test_struct_roundtrip(self) -> None:
        """Struct types can round-trip through JSON."""
        original = DataTypes.STRUCT(
            fields={"id": DataTypes.INT64, "value": DataTypes.FLOAT64}
        )
        deserialised = DataTypes.model_validate_json(original.value)
        assert deserialised.value == original.value
        assert deserialised.fields is not None
        assert len(deserialised.fields) == 2

    def test_nested_struct(self) -> None:
        """Structs can contain other Structs."""
        inner_struct = DataTypes.STRUCT(fields={"x": DataTypes.FLOAT64})
        outer_struct = DataTypes.STRUCT(fields={"point": inner_struct})

        deserialised = DataTypes.model_validate_json(outer_struct.value)
        assert deserialised.fields is not None
        assert "point" in deserialised.fields
        point_field = deserialised.fields["point"]
        assert point_field.base_type.value == "Struct"
        assert point_field.fields is not None
        assert "x" in point_field.fields

    def test_struct_with_array(self) -> None:
        """Structs can contain Arrays."""
        struct = DataTypes.STRUCT(
            fields={
                "name": DataTypes.STRING,
                "values": DataTypes.ARRAY(DataTypes.FLOAT64, shape=5),
            }
        )
        deserialised = DataTypes.model_validate_json(struct.value)
        assert deserialised.fields is not None
        values_field = deserialised.fields["values"]
        assert values_field.base_type.value == "Array"
        assert values_field.shape == (5,)

    def test_struct_with_list(self) -> None:
        """Structs can contain Lists."""
        struct = DataTypes.STRUCT(
            fields={
                "tags": DataTypes.LIST(DataTypes.STRING),
                "count": DataTypes.INT32,
            }
        )
        deserialised = DataTypes.model_validate_json(struct.value)
        assert deserialised.fields is not None
        tags_field = deserialised.fields["tags"]
        assert tags_field.base_type.value == "List"


class TestPolarsConversion:
    """Test conversion to/from Polars DataTypes."""

    @pytest.mark.parametrize(
        ("pl_dtype", "expected_base_type"),
        [
            (pl.String, "String"),
            (pl.Int64, "Int64"),
            (pl.Float64, "Float64"),
            (pl.Boolean, "Boolean"),
        ],
    )
    def test_from_polars_simple(
        self, pl_dtype: type[pl.DataType], expected_base_type: str
    ) -> None:
        """Convert simple Polars types to DataTypes."""
        dt = DataTypes.from_dtype(pl_dtype)
        assert dt.base_type.value == expected_base_type

    def test_from_polars_array(self) -> None:
        """Convert Polars Array to DataTypes."""
        pl_array = pl.Array(pl.Float32, 7)
        dt = DataTypes.from_dtype(pl_array)
        assert dt.base_type.value == "Array"
        assert dt.shape == (7,)
        assert dt.inner is not None
        assert dt.inner.base_type.value == "Float32"

    def test_from_polars_list(self) -> None:
        """Convert Polars List to DataTypes."""
        pl_list = pl.List(pl.Int64)
        dt = DataTypes.from_dtype(pl_list)
        assert dt.base_type.value == "List"
        assert dt.inner is not None
        assert dt.inner.base_type.value == "Int64"

    def test_from_polars_struct(self) -> None:
        """Convert Polars Struct to DataTypes."""
        pl_struct = pl.Struct([pl.Field("name", pl.String), pl.Field("age", pl.Int64)])
        dt = DataTypes.from_dtype(pl_struct)
        assert dt.base_type.value == "Struct"
        assert dt.fields is not None
        assert len(dt.fields) == 2
        assert dt.fields["name"].base_type.value == "String"
        assert dt.fields["age"].base_type.value == "Int64"

    def test_to_polars_list(self) -> None:
        """Convert DataTypes List to Polars."""
        dt = DataTypes.LIST(DataTypes.STRING)
        pl_dtype = dt.to_dtype()
        assert isinstance(pl_dtype, pl.List)

    def test_to_polars_array(self) -> None:
        """Convert DataTypes Array to Polars."""
        dt = DataTypes.ARRAY(DataTypes.INT64, shape=4)
        pl_dtype = dt.to_dtype()
        assert isinstance(pl_dtype, pl.Array)
        assert pl_dtype.shape == (4,)

    def test_to_polars_struct(self) -> None:
        """Convert DataTypes Struct to Polars."""
        dt = DataTypes.STRUCT(fields={"x": DataTypes.FLOAT64, "y": DataTypes.FLOAT64})
        pl_dtype = dt.to_dtype()
        assert isinstance(pl_dtype, pl.Struct)
        assert len(pl_dtype.fields) == 2

    def test_polars_roundtrip_complex(self) -> None:
        """Complex types can round-trip through Polars."""
        original_pl = pl.Struct(
            [
                pl.Field("id", pl.Int64),
                pl.Field("coords", pl.Array(pl.Float64, 2)),
                pl.Field("tags", pl.List(pl.String)),
            ]
        )
        dt = DataTypes.from_dtype(original_pl)
        back_to_pl = dt.to_dtype()

        assert isinstance(back_to_pl, pl.Struct)
        assert len(back_to_pl.fields) == 3


class TestComplexTypes:
    """Test complex nested type structures."""

    def test_list_of_structs(self) -> None:
        """List can contain Structs."""
        struct = DataTypes.STRUCT(fields={"a": DataTypes.INT64})
        list_of_structs = DataTypes.LIST(struct)

        deserialised = DataTypes.model_validate_json(list_of_structs.value)
        assert deserialised.inner is not None
        assert deserialised.inner.base_type.value == "Struct"

    def test_array_of_structs(self) -> None:
        """Array can contain Structs."""
        struct = DataTypes.STRUCT(fields={"value": DataTypes.FLOAT64})
        array_of_structs = DataTypes.ARRAY(struct, shape=3)

        deserialised = DataTypes.model_validate_json(array_of_structs.value)
        assert deserialised.shape == (3,)
        assert deserialised.inner is not None
        assert deserialised.inner.base_type.value == "Struct"

    def test_deeply_nested(self) -> None:
        """Multiple levels of nesting serialise correctly."""
        nested = DataTypes.STRUCT(
            fields={
                "metadata": DataTypes.STRUCT(
                    fields={"id": DataTypes.STRING, "version": DataTypes.INT32}
                ),
                "data": DataTypes.LIST(DataTypes.ARRAY(DataTypes.FLOAT64, shape=10)),
            }
        )

        deserialised = DataTypes.model_validate_json(nested.value)
        assert deserialised.fields is not None
        assert "metadata" in deserialised.fields
        assert "data" in deserialised.fields

        metadata = deserialised.fields["metadata"]
        assert metadata.fields is not None
        assert len(metadata.fields) == 2

        data = deserialised.fields["data"]
        assert data.inner is not None
        assert data.inner.shape == (10,)



from typing import Any

import numpy as np
import polars as pl
import pytest

from matchbox.common.arrow import SCHEMA_RESULTS
from matchbox.common.factories.entities import (
    ClusterEntity,
    EntityReference,
    SourceEntity,
)
from matchbox.common.factories.models import (
    calculate_min_max_edges,
    component_report,
    generate_dummy_probabilities,
    generate_entity_probabilities,
)
from matchbox.common.transform import DisjointSet
from test.common.factories.test_entity_factory import (
    make_cluster_entity,
    make_source_entity,
)


@pytest.mark.parametrize(
    ("left_n", "right_n", "n_components", "true_min", "true_max"),
    [
        (10, None, 2, 8, 20),
        (11, None, 2, 9, 25),
        (9, 9, 3, 15, 27),
        (8, 4, 3, 9, 11),
        (4, 8, 3, 9, 11),
        (8, 8, 3, 13, 22),
    ],
    ids=[
        "dedupe_no_mod",
        "dedup_mod",
        "link_no_mod",
        "link_left_mod",
        "link_right_mod",
        "link_same_mod",
    ],
)
def test_calculate_min_max_edges(
    left_n: int,
    right_n: int | None,
    n_components: int,
    true_min: int,
    true_max: int,
) -> None:
    deduplicate = False
    if not right_n:
        deduplicate = True
        right_n = left_n
    min_edges, max_edges = calculate_min_max_edges(
        left_n, right_n, n_components, deduplicate
    )

    assert true_min == min_edges
    assert true_max == max_edges


@pytest.mark.parametrize(
    ("parameters"),
    [
        pytest.param(
            {
                "left_count": 5,
                "right_count": None,
                "prob_range": (0.6, 0.8),
                "num_components": 3,
                "total_rows": 2,
            },
            id="dedupe_no_edges",
        ),
        pytest.param(
            {
                "left_count": 1_000,
                "right_count": None,
                "prob_range": (0.6, 0.8),
                "num_components": 10,
                "total_rows": calculate_min_max_edges(1_000, 1_000, 10, True)[0],
            },
            id="dedupe_min",
        ),
        pytest.param(
            {
                "left_count": 1_000,
                "right_count": None,
                "prob_range": (0.6, 0.8),
                "num_components": 10,
                "total_rows": calculate_min_max_edges(1_000, 1_000, 10, True)[1],
            },
            id="dedupe_max",
        ),
        pytest.param(
            {
                "left_count": 1_000,
                "right_count": 1_000,
                "prob_range": (0.6, 0.8),
                "num_components": 10,
                "total_rows": calculate_min_max_edges(1_000, 1_000, 10, False)[0],
            },
            id="link_min",
        ),
        pytest.param(
            {
                "left_count": 1_000,
                "right_count": 1_000,
                "prob_range": (0.6, 0.8),
                "num_components": 10,
                "total_rows": calculate_min_max_edges(1_000, 1_000, 10, False)[1],
            },
            id="link_max",
        ),
        pytest.param(
            {
                "left_count": 1_000,
                "right_count": 1_000,
                "prob_range": (0.6, 0.8),
                "num_components": 10,
                "total_rows": None,
            },
            id="blank_total_rows",
        ),
    ],
)
def test_generate_dummy_probabilities(parameters: dict[str, Any]) -> None:
    len_left = parameters["left_count"]
    len_right = parameters["right_count"]
    if len_right:
        total_len = len_left + len_right
        len_right = parameters["right_count"]
        rand_vals = np.random.choice(a=total_len, replace=False, size=total_len)
        left_values = tuple(rand_vals[:len_left])
        right_values = tuple(rand_vals[len_left:])
    else:
        rand_vals = np.random.choice(a=len_left, replace=False, size=len_left)
        left_values = tuple(rand_vals[:len_left])
        right_values = None

    n_components = parameters["num_components"]
    total_rows = parameters["total_rows"]
    min_edges, _ = calculate_min_max_edges(
        parameters["left_count"],
        parameters["right_count"] or parameters["left_count"],
        parameters["num_components"],
        right_values is None,
    )
    total_rows = total_rows or min_edges

    probabilities = generate_dummy_probabilities(
        left_values=left_values,
        right_values=right_values,
        prob_range=parameters["prob_range"],
        num_components=n_components,
        total_rows=total_rows,
    )
    report = component_report(table=probabilities, all_nodes=rand_vals)
    p_left = probabilities["left_id"].to_list()
    p_right = probabilities["right_id"].to_list()

    assert report["num_components"] == n_components

    # Link job
    if right_values:
        assert set(p_left) <= set(left_values)
        assert set(p_right) <= set(right_values)
    # Dedupe
    else:
        assert set(p_left) | set(p_right) <= set(left_values)

    assert probabilities["probability"].max() / 100 <= parameters["prob_range"][1]
    assert probabilities["probability"].min() / 100 >= parameters["prob_range"][0]

    assert len(probabilities) == total_rows

    edges = zip(p_left, p_right, strict=True)
    edges_set = {tuple(sorted(e)) for e in edges}
    assert len(edges_set) == total_rows

    self_references = [e for e in edges if e[0] == e[1]]
    assert len(self_references) == 0


def test_generate_dummy_probabilities_no_self_references() -> None:
    # Create input with repeated values
    left_values = tuple([1] * 4 + [2] * 4 + [3] * 4)

    try:
        probabilities = generate_dummy_probabilities(
            left_values=left_values,
            right_values=None,
            prob_range=(0.6, 0.8),
            num_components=3,
            total_rows=3,
        )
    except ValueError:
        return

    # If no ValueError was raised, continue with the rest of the checks
    p_left = probabilities["left_id"].to_list()
    p_right = probabilities["right_id"].to_list()

    # Check for self-references
    self_references = [
        (l_, r_) for l_, r_ in zip(p_left, p_right, strict=False) if l_ == r_
    ]
    assert len(self_references) == 0, f"Found self-references: {self_references}"


@pytest.mark.parametrize(
    ("parameters"),
    [
        {
            "left_range": (0, 10_000),
            "right_range": (10_000, 20_000),
            "num_components": 2,
            "total_rows": 1,
        },
        {
            "left_range": (0, 10),
            "right_range": (10, 20),
            "num_components": 2,
            "total_rows": 8_000,
        },
    ],
    ids=["lower_than_min", "higher_than_max"],
)
def test_generate_dummy_probabilities_errors(parameters: dict[str, Any]) -> None:
    left_values = tuple(range(*parameters["left_range"]))
    right_values = tuple(range(*parameters["right_range"]))

    with pytest.raises(ValueError):
        generate_dummy_probabilities(
            left_values=left_values,
            right_values=right_values,
            prob_range=(0.6, 0.8),
            num_components=parameters["num_components"],
            total_rows=parameters["total_rows"],
        )


@pytest.mark.parametrize(
    ("left_entities", "right_entities", "source_entities", "prob_range", "expected"),
    [
        pytest.param(
            frozenset(
                [
                    make_cluster_entity(1, "test", ["a1"]),
                    make_cluster_entity(2, "test", ["a2"]),
                ]
            ),
            None,  # Deduplication case
            frozenset([make_source_entity("test", ["a1", "a2"], "a")]),
            (0.8, 1.0),
            {"edge_count": 1, "prob_range": (80, 100)},
            id="basic_dedupe",
        ),
        pytest.param(
            frozenset([make_cluster_entity(1, "left", ["a1"])]),
            frozenset([make_cluster_entity(2, "right", ["b1"])]),
            frozenset(
                [
                    make_source_entity("left", ["a1"], "a"),
                    make_source_entity("right", ["b1"], "b"),
                ]
            ),
            (0.8, 1.0),
            {"edge_count": 0, "prob_range": (80, 100)},
            id="basic_link_no_match",
        ),
        pytest.param(
            frozenset([make_cluster_entity(1, "test", ["a1"])]),
            frozenset([make_cluster_entity(2, "test", ["a2"])]),
            frozenset([make_source_entity("test", ["a1", "a2"], "a")]),
            (0.8, 1.0),
            {"edge_count": 1, "prob_range": (80, 100)},
            id="successful_link",
        ),
        pytest.param(
            frozenset(
                [
                    make_cluster_entity(1, "test", ["a1", "a2"]),
                    make_cluster_entity(2, "test", ["a2", "a3"]),
                    make_cluster_entity(3, "test", ["a3", "a4"]),
                ]
            ),
            None,
            frozenset(
                [make_source_entity("test", ["a1", "a2", "a3", "a4"], "entity_a")]
            ),
            (0.8, 1.0),
            {"edge_count": 3, "prob_range": (80, 100)},
            id="overlapping_dedupe",
        ),
        pytest.param(
            frozenset(
                [
                    make_cluster_entity(1, "test", ["a1"]),
                    make_cluster_entity(2, "test", ["b1"]),
                ]
            ),
            frozenset(
                [
                    make_cluster_entity(3, "test", ["a2"]),
                    make_cluster_entity(4, "test", ["b2"]),
                ]
            ),
            frozenset(
                [
                    make_source_entity("test", ["a1", "a2"], "a"),
                    make_source_entity("test", ["b1", "b2"], "b"),
                ]
            ),
            (0.8, 1.0),
            {"edge_count": 2, "prob_range": (80, 100)},
            id="multi_component_link",
        ),
        pytest.param(
            frozenset(
                [
                    make_cluster_entity(1, "test", ["a1"]),
                    make_cluster_entity(2, "test", ["a2"]),
                    make_cluster_entity(3, "test", ["x1"]),  # No source for this
                    make_cluster_entity(4, "test", ["y1"]),  # No source for this
                ]
            ),
            None,
            frozenset([make_source_entity("test", ["a1", "a2"], "a")]),
            (0.8, 1.0),
            {"edge_count": 1, "prob_range": (80, 100)},
            id="partial_source_coverage",
        ),
        pytest.param(
            frozenset(),
            frozenset(),
            frozenset(),
            (0.8, 1.0),
            {"edge_count": 0, "prob_range": (80, 100)},
            id="empty_sets",
        ),
        pytest.param(
            frozenset(
                [
                    make_cluster_entity(1, "test", ["a1"]),
                    make_cluster_entity(2, "test", ["a2"]),
                ]
            ),
            None,
            frozenset([make_source_entity("test", ["a1", "a2"], "a")]),
            (0.5, 0.7),
            {"edge_count": 1, "prob_range": (50, 70)},
            id="different_prob_range",
        ),
        pytest.param(
            frozenset(
                [
                    make_cluster_entity(1, "test", ["a1", "a2"]),  # Merged
                    make_cluster_entity(2, "test", ["a3"]),  # Unmerged
                    make_cluster_entity(3, "test", ["b1"]),  # From different source
                ]
            ),
            None,
            frozenset(
                [
                    make_source_entity("test", ["a1", "a2", "a3"], "a"),
                    make_source_entity("test", ["b1"], "b"),
                ]
            ),
            (0.8, 1.0),
            {"edge_count": 1, "prob_range": (80, 100)},
            id="mixed_merged_unmerged",
        ),
        pytest.param(
            frozenset(
                [
                    make_cluster_entity(1, "source1", ["1"]),
                    make_cluster_entity(2, "source2", ["A"]),
                    make_cluster_entity(3, "source3", ["X"]),
                ]
            ),
            None,
            frozenset(
                [
                    make_source_entity("source1", ["1"], "entity"),
                    make_source_entity("source2", ["A"], "entity"),
                    make_source_entity("source3", ["X"], "entity"),
                ]
            ),
            (0.8, 1.0),
            {"edge_count": 0, "prob_range": (80, 100)},
            id="multi_source_entity",
        ),
    ],
)
def test_generate_entity_probabilities_scenarios(
    left_entities: frozenset[ClusterEntity],
    right_entities: frozenset[ClusterEntity] | None,
    source_entities: frozenset[SourceEntity],
    prob_range: tuple[float, float],
    expected: dict,
) -> None:
    """Comprehensive test for generate_entity_probabilities with various scenarios."""
    # Run the function
    result = generate_entity_probabilities(
        left_entities, right_entities, source_entities, prob_range
    )

    # Check schema
    assert result.schema == pl.Schema(SCHEMA_RESULTS)

    # Get edges from result
    edges = list(
        zip(
            result["left_id"].to_list(),
            result["right_id"].to_list(),
            strict=True,
        )
    )

    # Check number of edges matches expected
    assert len(edges) == expected["edge_count"]

    # For non-empty results, validate probability ranges
    if edges:
        probs = result["probability"].to_numpy()
        prob_min, prob_max = expected["prob_range"]
        assert all(prob_min <= p <= prob_max for p in probs)


@pytest.mark.parametrize(
    ("seed1", "seed2", "should_be_equal", "case"),
    [
        pytest.param(42, 42, True, "dedupe", id="same_seeds_dedupe"),
        pytest.param(1, 2, False, "dedupe", id="different_seeds_dedupe"),
        pytest.param(42, 42, True, "link", id="same_seeds_link"),
        pytest.param(1, 2, False, "link", id="different_seeds_link"),
    ],
)
def test_seed_determinism(
    seed1: int,
    seed2: int,
    should_be_equal: bool,
    case: str,
) -> None:
    """Test that seeds produce consistent/different results as expected."""
    # Create test entities
    source = make_source_entity("test", ["a1", "a2", "a3"], "entity_a")
    entities = frozenset(
        [
            make_cluster_entity(1, "test", ["a1"]),
            make_cluster_entity(2, "test", ["a2"]),
            make_cluster_entity(3, "test", ["a3"]),
        ]
    )

    if case == "dedupe":
        right_entities = None
    else:
        # For linking case, use second set of entities
        right_entities = frozenset(
            [
                make_cluster_entity(4, "test", ["a1"]),
                make_cluster_entity(5, "test", ["a2"]),
                make_cluster_entity(6, "test", ["a3"]),
            ]
        )

    # Generate results with the two seeds
    result1 = generate_entity_probabilities(
        left_entities=entities,
        right_entities=right_entities,
        source_entities=frozenset([source]),
        prob_range=(0.8, 1.0),
        seed=seed1,
    )

    result2 = generate_entity_probabilities(
        left_entities=entities,
        right_entities=right_entities,
        source_entities=frozenset([source]),
        prob_range=(0.8, 1.0),
        seed=seed2,
    )

    assert result1.shape[0] > 0
    assert result2.shape[0] > 0

    if should_be_equal:
        assert result1.equals(result2)
    else:
        assert not result1.equals(result2)


def test_disjoint_set_recovery() -> None:
    """Test that DisjointSet can recover the entity structure from probabilities."""
    # Create source entities
    source1 = make_source_entity("source1", ["1", "2", "3"], "entity1")
    source2 = make_source_entity("source1", ["4", "5", "6"], "entity2")

    # Create split cluster entities
    clusters = frozenset(
        [
            make_cluster_entity(1, "source1", ["1"]),
            make_cluster_entity(2, "source1", ["2"]),
            make_cluster_entity(3, "source1", ["3"]),
            make_cluster_entity(4, "source1", ["4"]),
            make_cluster_entity(5, "source1", ["5"]),
            make_cluster_entity(6, "source1", ["6"]),
        ]
    )

    # Generate probabilities
    table = generate_entity_probabilities(
        left_entities=clusters,
        right_entities=None,
        source_entities=frozenset([source1, source2]),
        prob_range=(0.9, 1.0),
    )

    # Use DisjointSet to cluster based on high probabilities
    ds = DisjointSet[int]()
    for row in table.to_dicts():
        if row["probability"] >= 90:  # High confidence matches
            ds.union(row["left_id"], row["right_id"])

    # Get resulting clusters
    clusters = ds.get_components()

    # Should recover original entities - exactly two clusters
    assert len(clusters) == 2

    # Each cluster should contain the right number of ClusterEntity objects
    cluster_sizes = sorted(len(cluster) for cluster in clusters)
    assert cluster_sizes == [3, 3]


@pytest.mark.parametrize(
    "prob_range",
    [
        pytest.param((-0.1, 0.5), id="negative_lower_bound"),  # Negative lower bound
        pytest.param((0.5, 1.1), id="upper_bound_too_high"),  # Upper bound > 1.0
        pytest.param((0.8, 0.7), id="decreasing_range"),  # Decreasing range
    ],
)
def test_invalid_probability_ranges(prob_range: tuple[float, float]) -> None:
    """Test that invalid probability ranges raise appropriate errors."""
    source = make_source_entity("test", ["a1", "a2"], "entity")
    entities = frozenset(
        [
            make_cluster_entity(1, "test", ["a1"]),
            make_cluster_entity(2, "test", ["a2"]),
        ]
    )

    with pytest.raises(ValueError, match="Probabilities must be"):
        generate_entity_probabilities(
            left_entities=entities,
            right_entities=None,
            source_entities=frozenset([source]),
            prob_range=prob_range,
        )


def test_complex_entity_recovery() -> None:
    """Test recovery of complex, multi-source entity structures."""
    # Create a source entity spanning multiple sources
    source = SourceEntity(
        base_values={"name": "Complex Entity"},
        keys=EntityReference(
            {
                "source1": frozenset(["1", "2"]),
                "source2": frozenset(["A", "B"]),
                "source3": frozenset(["X"]),
            }
        ),
    )

    # Create fragmented ClusterEntity objects
    clusters = frozenset(
        [
            ClusterEntity(keys=EntityReference({"source1": frozenset(["1"])})),
            ClusterEntity(keys=EntityReference({"source1": frozenset(["2"])})),
            ClusterEntity(keys=EntityReference({"source2": frozenset(["A"])})),
            ClusterEntity(keys=EntityReference({"source2": frozenset(["B"])})),
            ClusterEntity(keys=EntityReference({"source3": frozenset(["X"])})),
        ]
    )

    # Generate probabilities
    table = generate_entity_probabilities(
        left_entities=clusters,
        right_entities=None,
        source_entities=frozenset([source]),
        prob_range=(0.9, 1.0),
    )

    # There should be edges connecting all entities (n*(n-1))/2 = 10 edges
    assert len(table) == 10

    # Use DisjointSet to cluster
    ds = DisjointSet[int]()
    for row in table.to_dicts():
        if row["probability"] >= 90:
            ds.union(row["left_id"], row["right_id"])

    # Should recover as a single component
    clusters = ds.get_components()
    assert len(clusters) == 1
    assert len(clusters[0]) == 5  # All entities in one cluster



"""Admin PostgreSQL mixin for Matchbox server."""

from typing import Literal

from sqlalchemy import and_, bindparam, delete, select, union_all

from matchbox.common.dtos import (
    BackendResourceType,
    CollectionName,
    Group,
    GroupName,
    PermissionGrant,
    PermissionType,
    User,
)
from matchbox.common.exceptions import (
    MatchboxCollectionNotFoundError,
    MatchboxDataNotFound,
    MatchboxDeletionNotConfirmed,
    MatchboxGroupAlreadyExistsError,
    MatchboxGroupNotFoundError,
    MatchboxSystemGroupError,
    MatchboxUserNotFoundError,
)
from matchbox.common.logging import logger
from matchbox.server.base import PERMISSION_GRANTS, MatchboxSnapshot
from matchbox.server.postgresql.db import MBDB, MatchboxBackends
from matchbox.server.postgresql.orm import (
    Clusters,
    ClusterSourceKey,
    Collections,
    EvalJudgements,
    Groups,
    Permissions,
    Probabilities,
    Results,
    UserGroups,
    Users,
)
from matchbox.server.postgresql.utils.db import dump, restore


class MatchboxPostgresAdminMixin:
    """Admin mixin for the PostgreSQL adapter for Matchbox."""

    # User management

    def login(self, user: User) -> User:  # noqa: D102
        with MBDB.get_session() as session:
            # Try to find existing user
            existing_user = session.scalar(
                select(Users).where(Users.name == user.user_name)
            )

            if existing_user:
                # Update email if provided
                if user.email and existing_user.email != user.email:
                    existing_user.email = user.email
                    session.commit()

                return User(
                    user_name=existing_user.name,
                    email=existing_user.email,
                )

            # Create new user
            new_user = Users(name=user.user_name, email=user.email)
            session.add(new_user)
            session.commit()

            return User(
                user_name=new_user.name,
                email=new_user.email,
            )

    # Group management

    def get_user_groups(self, user_name: str) -> list[GroupName]:  # noqa: D102
        with MBDB.get_session() as session:
            user = session.scalar(select(Users).where(Users.name == user_name))

            if not user:
                raise MatchboxUserNotFoundError(f"User '{user_name}' not found")

            group_names = [GroupName(group.name) for group in user.groups]

            return group_names

    def list_groups(self) -> list[Group]:  # noqa: D102
        with MBDB.get_session() as session:
            groups_orm = session.scalars(select(Groups)).all()

            groups = [
                Group(
                    name=GroupName(g.name),
                    description=g.description,
                    is_system=g.is_system,
                    members=[
                        User(
                            user_id=member.user_id,
                            user_name=member.name,
                            email=member.email,
                        )
                        for member in g.members
                    ],
                )
                for g in groups_orm
            ]

            return groups

    def get_group(self, name: GroupName) -> Group:  # noqa: D102
        with MBDB.get_session() as session:
            group = session.scalar(select(Groups).where(Groups.name == name))

            if not group:
                raise MatchboxGroupNotFoundError(f"Group '{name}' not found")

            return Group(
                name=GroupName(group.name),
                description=group.description,
                is_system=group.is_system,
                members=[
                    User(
                        user_id=member.user_id,
                        user_name=member.name,
                        email=member.email,
                    )
                    for member in group.members
                ],
            )

    def create_group(self, group: Group) -> None:  # noqa: D102
        with MBDB.get_session() as session:
            # Check if group already exists
            existing = session.scalar(select(Groups).where(Groups.name == group.name))

            if existing:
                raise MatchboxGroupAlreadyExistsError(
                    f"Group '{group.name}' already exists"
                )

            # Create the group
            new_group = Groups(
                name=group.name,
                description=group.description,
                is_system=group.is_system,
            )
            session.add(new_group)
            session.commit()

            logger.info(f"Created group '{group.name}'", prefix="Create group")

    def delete_group(self, name: GroupName, certain: bool = False) -> None:  # noqa: D102
        if not certain:
            raise MatchboxDeletionNotConfirmed(
                f"This operation will delete the group '{name}' and all its "
                "permission grants. If you're sure you want to continue, rerun "
                "with certain=True"
            )

        with MBDB.get_session() as session:
            group = session.scalar(select(Groups).where(Groups.name == name))

            if not group:
                raise MatchboxGroupNotFoundError(f"Group '{name}' not found")

            if group.is_system:
                raise MatchboxSystemGroupError(f"Cannot delete system group '{name}'")

            session.delete(group)
            session.commit()

            logger.info(f"Deleted group '{name}'", prefix="Delete group")

    # User-group membership

    def add_user_to_group(self, user_name: str, group_name: GroupName) -> None:  # noqa: D102
        with MBDB.get_session() as session:
            # Get user
            user = session.scalar(select(Users).where(Users.name == user_name))
            if not user:
                raise MatchboxUserNotFoundError(f"User '{user_name}' not found")

            # Get group
            group = session.scalar(select(Groups).where(Groups.name == group_name))
            if not group:
                raise MatchboxGroupNotFoundError(f"Group '{group_name}' not found")

            # Check if already a member
            existing_membership = session.scalar(
                select(UserGroups).where(
                    and_(
                        UserGroups.user_id == user.user_id,
                        UserGroups.group_id == group.group_id,
                    )
                )
            )

            if not existing_membership:
                membership = UserGroups(
                    user_id=user.user_id,
                    group_id=group.group_id,
                )
                session.add(membership)

                session.commit()

                logger.info(
                    f"Added user '{user_name}' to group '{group_name}'",
                    prefix="Add user to group",
                )
            else:
                logger.info(
                    f"User '{user_name}' already belongs to '{group_name}'",
                    prefix="Add user to group",
                )

    def remove_user_from_group(self, user_name: str, group_name: GroupName) -> None:  # noqa: D102
        with MBDB.get_session() as session:
            # Get user
            user = session.scalar(select(Users).where(Users.name == user_name))
            if not user:
                raise MatchboxUserNotFoundError(f"User '{user_name}' not found")

            # Get group
            group = session.scalar(select(Groups).where(Groups.name == group_name))
            if not group:
                raise MatchboxGroupNotFoundError(f"Group '{group_name}' not found")

            # Delete membership
            result = session.execute(
                delete(UserGroups).where(
                    and_(
                        UserGroups.user_id == user.user_id,
                        UserGroups.group_id == group.group_id,
                    )
                )
            )

            session.commit()

            if result.rowcount > 0:
                logger.info(
                    f"Removed user '{user_name}' from group '{group_name}'",
                    prefix="Remove user from group",
                )
            else:
                logger.info(
                    f"User '{user_name}' not present in group '{group_name}'",
                    prefix="Remove user from group",
                )

    # Permissions management

    def check_permission(  # noqa: D102
        self,
        user_name: str,
        permission: PermissionType,
        resource: Literal[BackendResourceType.SYSTEM] | CollectionName,
    ) -> bool:
        with MBDB.get_session() as session:
            # Get user
            user = session.scalar(select(Users).where(Users.name == user_name))
            if not user:
                return False

            # Get user's group IDs
            user_group_ids = session.scalars(
                select(UserGroups.group_id).where(UserGroups.user_id == user.user_id)
            ).all()

            if not user_group_ids:
                return False

            # Get permissions that would satisfy this check
            sufficient_permissions = PERMISSION_GRANTS[permission]

            # Check permissions based on resource type
            if resource == BackendResourceType.SYSTEM:
                # Check system permissions
                grant = session.scalar(
                    select(Permissions).where(
                        and_(
                            Permissions.is_system == True,  # noqa: E712
                            Permissions.group_id.in_(user_group_ids),
                            Permissions.permission.in_(sufficient_permissions),
                        )
                    )
                )
            else:
                # Check collection permissions
                collection = session.scalar(
                    select(Collections).where(Collections.name == resource)
                )
                if not collection:
                    raise MatchboxCollectionNotFoundError(name=resource)

                grant = session.scalar(
                    select(Permissions).where(
                        and_(
                            Permissions.collection_id == collection.collection_id,
                            Permissions.group_id.in_(user_group_ids),
                            Permissions.permission.in_(sufficient_permissions),
                        )
                    )
                )

            return grant is not None

    def get_permissions(  # noqa: D102
        self,
        resource: Literal[BackendResourceType.SYSTEM] | CollectionName,
    ) -> list[PermissionGrant]:
        with MBDB.get_session() as session:
            if resource == BackendResourceType.SYSTEM:
                # Get system permissions
                permissions_query = (
                    select(Permissions)
                    .where(Permissions.is_system == True)  # noqa: E712
                    .join(Groups, Permissions.group_id == Groups.group_id)
                )
            else:
                # Get collection permissions
                collection = session.scalar(
                    select(Collections).where(Collections.name == resource)
                )
                if not collection:
                    raise MatchboxCollectionNotFoundError(name=resource)

                permissions_query = (
                    select(Permissions)
                    .where(Permissions.collection_id == collection.collection_id)
                    .join(Groups, Permissions.group_id == Groups.group_id)
                )

            permissions_orm = session.scalars(permissions_query).all()

            grants = [
                PermissionGrant(
                    group_name=GroupName(perm.group.name),
                    permission=PermissionType(perm.permission),
                )
                for perm in permissions_orm
            ]

            return grants

    def grant_permission(  # noqa: D102
        self,
        group_name: GroupName,
        permission: PermissionType,
        resource: Literal[BackendResourceType.SYSTEM] | CollectionName,
    ) -> None:
        with MBDB.get_session() as session:
            # Get group
            group = session.scalar(select(Groups).where(Groups.name == group_name))
            if not group:
                raise MatchboxGroupNotFoundError(f"Group '{group_name}' not found")

            if resource == BackendResourceType.SYSTEM:
                # Grant system permission
                # Check if already exists
                existing = session.scalar(
                    select(Permissions).where(
                        and_(
                            Permissions.group_id == group.group_id,
                            Permissions.permission == permission,
                            Permissions.is_system == True,  # noqa: E712
                        )
                    )
                )

                if not existing:
                    new_permission = Permissions(
                        group_id=group.group_id,
                        permission=permission,
                        is_system=True,
                    )
                    session.add(new_permission)
            else:
                # Grant collection permission
                collection = session.scalar(
                    select(Collections).where(Collections.name == resource)
                )
                if not collection:
                    raise MatchboxCollectionNotFoundError(name=resource)

                # Check if already exists
                existing = session.scalar(
                    select(Permissions).where(
                        and_(
                            Permissions.group_id == group.group_id,
                            Permissions.permission == permission,
                            Permissions.collection_id == collection.collection_id,
                        )
                    )
                )

                if not existing:
                    new_permission = Permissions(
                        group_id=group.group_id,
                        permission=permission,
                        collection_id=collection.collection_id,
                    )
                    session.add(new_permission)

            session.commit()

            logger.info(
                f"Granted {permission} permission on '{resource}' "
                f"to group '{group_name}'",
                prefix="Grant permission",
            )

    def revoke_permission(  # noqa: D102
        self,
        group_name: GroupName,
        permission: PermissionType,
        resource: Literal[BackendResourceType.SYSTEM] | CollectionName,
    ) -> None:
        with MBDB.get_session() as session:
            # Get group
            group = session.scalar(select(Groups).where(Groups.name == group_name))
            if not group:
                raise MatchboxGroupNotFoundError(f"Group '{group_name}' not found")

            if resource == BackendResourceType.SYSTEM:
                # Revoke system permission
                result = session.execute(
                    delete(Permissions).where(
                        and_(
                            Permissions.group_id == group.group_id,
                            Permissions.permission == permission,
                            Permissions.is_system == True,  # noqa: E712
                        )
                    )
                )
            else:
                # Revoke collection permission
                collection = session.scalar(
                    select(Collections).where(Collections.name == resource)
                )
                if not collection:
                    raise MatchboxCollectionNotFoundError(name=resource)

                result = session.execute(
                    delete(Permissions).where(
                        and_(
                            Permissions.group_id == group.group_id,
                            Permissions.permission == permission,
                            Permissions.collection_id == collection.collection_id,
                        )
                    )
                )

            session.commit()

            if result.rowcount > 0:
                logger.info(
                    f"Revoked {permission} permission on '{resource}' "
                    f"from group '{group_name}'",
                    prefix="Revoke permission",
                )
            else:
                logger.info(
                    f"Permission {permission} on '{resource}' "
                    f"not present for group '{group_name}'",
                    prefix="Revoke permission",
                )

    # Data management

    def validate_ids(self, ids: list[int]) -> bool:  # noqa: D102
        with MBDB.get_session() as session:
            data_inner_join = (
                session.query(Clusters)
                .filter(
                    Clusters.cluster_id.in_(
                        bindparam(
                            "ins_ids",
                            ids,
                            expanding=True,
                        )
                    )
                )
                .all()
            )

        existing_ids = {item.cluster_id for item in data_inner_join}
        missing_ids = set(ids) - existing_ids

        if missing_ids:
            raise MatchboxDataNotFound(
                message="Some items don't exist in Clusters table.",
                table=Clusters.__tablename__,
                data=missing_ids,
            )

        return True

    def dump(self) -> MatchboxSnapshot:  # noqa: D102
        return dump()

    def drop(self, certain: bool) -> None:  # noqa: D102
        if certain:
            MBDB.drop_database()
        else:
            raise MatchboxDeletionNotConfirmed(
                "This operation will drop the entire database and recreate it."
                "It's not expected to be used as part normal operations."
                "If you're sure you want to continue, rerun with certain=True"
            )

    def clear(self, certain: bool) -> None:  # noqa: D102
        if certain:
            MBDB.clear_database()
        else:
            raise MatchboxDeletionNotConfirmed(
                "This operation will drop all rows in the database but not the "
                "tables themselves. It's primarily used to reset following tests."
                "If you're sure you want to continue, rerun with certain=True"
            )

    def restore(self, snapshot: MatchboxSnapshot) -> None:  # noqa: D102
        if snapshot.backend_type != MatchboxBackends.POSTGRES:
            raise TypeError(
                f"Cannot restore {snapshot.backend_type} snapshot to PostgreSQL backend"
            )

        MBDB.clear_database()

        restore(
            snapshot=snapshot,
            batch_size=self.settings.batch_size,
        )

    def delete_orphans(self) -> int:  # noqa: D102
        with MBDB.get_session() as session:
            # Get all cluster ids in related tables
            union_all_cte = union_all(
                select(EvalJudgements.endorsed_cluster_id.label("cluster_id")),
                select(EvalJudgements.shown_cluster_id.label("cluster_id")),
                select(ClusterSourceKey.cluster_id),
                select(Probabilities.cluster_id),
                select(Results.left_id.label("cluster_id")),
                select(Results.right_id.label("cluster_id")),
            ).cte("union_all_cte")

            # Deduplicate only once
            not_orphans = (
                select(union_all_cte.c.cluster_id).distinct().cte("not_orphans")
            )

            # Return clusters not in related tables
            stmt = delete(Clusters).where(
                ~select(not_orphans.c.cluster_id)
                .where(not_orphans.c.cluster_id == Clusters.cluster_id)
                .exists()
            )
            # Delete orphans
            deletion = session.execute(stmt)

            session.commit()

            logger.info(f"Deleted {deletion.rowcount} orphans", prefix="Delete orphans")
            return deletion.rowcount



"""Add user model, groups and permissions to ORM.

Revision ID: 13095b44ff09
Revises: c774fd4b69f8
Create Date: 2025-12-02 12:48:16.444413

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "13095b44ff09"
down_revision: str | None = "c774fd4b69f8"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    op.create_table(
        "groups",
        sa.Column("group_id", sa.BIGINT(), autoincrement=True, nullable=False),
        sa.Column("name", sa.TEXT(), nullable=False),
        sa.Column("description", sa.TEXT(), nullable=True),
        sa.Column("is_system", sa.BOOLEAN(), nullable=False),
        sa.PrimaryKeyConstraint("group_id"),
        sa.UniqueConstraint("name", name="groups_name_key"),
        schema="mb",
    )
    op.create_table(
        "permissions",
        sa.Column("permission_id", sa.BIGINT(), autoincrement=True, nullable=False),
        sa.Column("permission", sa.TEXT(), nullable=False),
        sa.Column("group_id", sa.BIGINT(), nullable=False),
        sa.Column("collection_id", sa.BIGINT(), nullable=True),
        sa.Column("is_system", sa.BOOLEAN(), nullable=True),
        sa.CheckConstraint(
            "permission IN ('read', 'write', 'admin')", name="valid_permission"
        ),
        sa.CheckConstraint(
            "(collection_id IS NOT NULL AND is_system IS NULL) "
            "OR (collection_id IS NULL AND is_system = true)",
            name="exactly_one_resource",
        ),
        sa.ForeignKeyConstraint(
            ["collection_id"], ["mb.collections.collection_id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["group_id"], ["mb.groups.group_id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("permission_id"),
        sa.UniqueConstraint(
            "permission",
            "group_id",
            "collection_id",
            "is_system",
            name="unique_permission_grant",
        ),
        schema="mb",
    )
    op.create_table(
        "user_groups",
        sa.Column("user_id", sa.BIGINT(), nullable=False),
        sa.Column("group_id", sa.BIGINT(), nullable=False),
        sa.ForeignKeyConstraint(
            ["group_id"], ["mb.groups.group_id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(["user_id"], ["mb.users.user_id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("user_id", "group_id"),
        schema="mb",
    )
    op.add_column("users", sa.Column("email", sa.TEXT(), nullable=True), schema="mb")


def downgrade() -> None:
    """Downgrade schema."""
    op.drop_column("users", "email", schema="mb")
    op.drop_table("user_groups", schema="mb")
    op.drop_table("permissions", schema="mb")
    op.drop_table("groups", schema="mb")



"""Integration tests with real scenario data - comprehensive behaviour testing."""

from collections.abc import Callable
from functools import partial
from unittest.mock import Mock

import pytest
from sqlalchemy import Engine
from textual.widgets import Footer, Label

from matchbox.client.cli.eval.app import EntityResolutionApp, EvaluationQueue
from matchbox.client.cli.eval.widgets.table import ComparisonDisplayTable
from matchbox.client.dags import DAG
from matchbox.common.dtos import ModelResolutionPath
from matchbox.common.factories.dags import TestkitDAG
from matchbox.common.factories.scenarios import setup_scenario
from matchbox.server.base import MatchboxDBAdapter

backends = [
    pytest.param("matchbox_postgres", id="postgres"),
]


@pytest.fixture(scope="function")
def backend_instance(request: pytest.FixtureRequest, backend: str) -> MatchboxDBAdapter:
    """Create a fresh backend instance for each test."""
    backend_obj = request.getfixturevalue(backend)
    backend_obj.clear(certain=True)
    return backend_obj


class TestEvaluationQueue:
    """Unit tests for EvaluationQueue - no app or mocking needed."""

    def test_queue_initialises_empty(self) -> None:
        """Test queue starts empty."""
        queue = EvaluationQueue()

        assert queue.total_count == 0
        assert queue.current is None

    def test_add_items_increases_count(self) -> None:
        """Test adding items increases count."""
        queue = EvaluationQueue()
        items = [Mock(item=Mock(leaves=[1, 2])), Mock(item=Mock(leaves=[3, 4]))]

        added = queue.add_sessions(items)

        assert added == 2
        assert queue.total_count == 2
        assert queue.current is items[0]

    def test_add_items_prevents_duplicates(self) -> None:
        """Test that duplicate cluster IDs are not added."""
        queue = EvaluationQueue()
        item1 = Mock(item=Mock(leaves=[1, 2]))
        item2 = Mock(item=Mock(leaves=[2, 1]))  # Duplicate

        queue.add_sessions([item1])
        added = queue.add_sessions([item2])

        assert added == 0
        assert queue.total_count == 1

    def test_add_items_handles_empty_list(self) -> None:
        """Test that adding empty list returns 0."""
        queue = EvaluationQueue()

        added = queue.add_sessions([])

        assert added == 0
        assert queue.total_count == 0

    def test_skip_rotates_deque(self) -> None:
        """Test that skip moves current to back."""
        queue = EvaluationQueue()
        item1 = Mock(item=Mock(cluster_id=1))
        item2 = Mock(item=Mock(cluster_id=2))
        queue.sessions.extend([item1, item2])

        queue.skip_current()

        assert queue.current is item2
        assert queue.sessions[1] is item1

    def test_skip_with_single_item_does_nothing(self) -> None:
        """Test that skip with one item doesn't rotate."""
        queue = EvaluationQueue()
        item1 = Mock(item=Mock(cluster_id=1))
        queue.sessions.append(item1)

        queue.skip_current()

        assert queue.current is item1
        assert queue.total_count == 1

    def test_remove_current_pops_front(self) -> None:
        """Test that remove_current removes from front."""
        queue = EvaluationQueue()
        item1 = Mock(item=Mock(cluster_id=1))
        item2 = Mock(item=Mock(cluster_id=2))
        queue.sessions.extend([item1, item2])

        removed = queue.remove_current()

        assert removed is item1
        assert queue.total_count == 1
        assert queue.current is item2

    def test_remove_current_on_empty_returns_none(self) -> None:
        """Test that remove_current on empty queue returns None."""
        queue = EvaluationQueue()

        removed = queue.remove_current()

        assert removed is None
        assert queue.total_count == 0


@pytest.mark.parametrize("backend", backends)
@pytest.mark.docker
class TestScenarioIntegration:
    """Integration tests using real scenario data with backend."""

    @pytest.fixture
    def test_resolution(self) -> ModelResolutionPath:
        """Create test resolution path."""
        return ModelResolutionPath(
            collection="test_collection", run=1, name="test_resolution"
        )

    @pytest.fixture(autouse=True)
    def setup(self, backend_instance: str, sqla_sqlite_warehouse: Engine) -> None:
        """Set up test fixtures."""
        self.backend: MatchboxDBAdapter = backend_instance
        self.warehouse_engine: Engine = sqla_sqlite_warehouse
        self.scenario: Callable[..., TestkitDAG] = partial(
            setup_scenario, warehouse=sqla_sqlite_warehouse
        )

    @pytest.mark.asyncio
    async def test_complete_evaluation_workflow(self) -> None:
        """Test complete evaluation workflow: start, load, label, clear, skip, submit.

        This comprehensive test covers the main user journey through the evaluation app:

        - App initialisation and sample loading
        - Keyboard-driven assignment workflow (letter → digit)
        - Clearing assignments
        - Skipping items in the queue
        - Submitting incomplete assignments (warning)
        - Completing and submitting assignments (saves judgement)
        - Status updates and help modal
        """
        with self.scenario(self.backend, "mega") as dag:
            model_name: str = "mega_product_linker"

            loaded_dag: DAG = dag.dag.load_pending().set_client(self.warehouse_engine)

            app = EntityResolutionApp(
                resolution=loaded_dag.get_model(model_name),
                num_samples=3,
                user="test_user",
                dag=loaded_dag,
                scroll_debounce_delay=None,
            )

            # Resolution carefully chosen so all columns on a page are always numbered
            async with app.run_test(size=(250, 150)) as pilot:
                await pilot.pause()

                # 1. Verify app initialisation and sample loading
                assert app.is_running
                assert app.user_name == "test_user"
                assert pilot.app.query("Footer")
                assert app.queue.total_count > 0
                assert app.queue.current is not None
                assert app.queue.current.item.leaves is not None

                # 2. Test keyboard workflow: letter then digit
                await pilot.press("a")
                await pilot.pause()
                table = app.query_one(ComparisonDisplayTable)
                assert table.current_group == "a"

                await pilot.press("1")
                await pilot.pause()
                current = app.queue.current
                assert current is not None
                assert 0 in current.assignments
                assert current.assignments[0] == "a"

                # 3. Verify assignment was saved
                assert len(current.assignments) == 1
                assert 0 in current.assignments

                # 3b. Test paging - scroll right and label different column
                initial_col_idx = table.cursor_column
                initial_scroll_x = table.scroll_x
                await pilot.press("right")
                await pilot.pause()

                # Verify paging behaviour
                assert table.scroll_x > initial_scroll_x, "Paging right should scroll"
                assert table.cursor_column != initial_col_idx, (
                    "Paging right should move cursor/headers"
                )

                # We pressed right, so '1' should now map to a new column
                new_cursor_idx = table.cursor_column
                await pilot.press("b")
                await pilot.press("1")
                await pilot.pause()
                expected_assignment_idx = new_cursor_idx - 1  # Cursor is 0-indexed
                assert expected_assignment_idx in current.assignments
                assert current.assignments[expected_assignment_idx] == "b"

                # Scroll back to start
                await pilot.press("left")
                await pilot.pause()
                assert table.scroll_x == initial_scroll_x

                # 4. Test status updates reactively
                assert isinstance(app.status, tuple)
                assert len(app.status) == 2

                # 5. Test clearing assignments
                await pilot.press("escape")
                await pilot.pause()
                assert len(current.assignments) == 0

                # 6. Test skip workflow
                first_item = app.queue.current
                await pilot.press("shift+right")
                await pilot.pause()
                assert app.queue.current is not first_item
                assert app.queue.sessions[-1] is first_item

                # 7. Test submitting incomplete assignment shows warning
                initial_count = app.queue.total_count
                await pilot.press("space")
                await pilot.pause()
                assert app.queue.total_count == initial_count
                status_message = app.status[0].lower()
                assert "incomplete" in status_message

                # 8. Test completing all columns via paging and successful submission
                current = app.queue.current
                assert current is not None

                initial_judgements, _ = self.backend.get_judgements()
                initial_judgement_count = len(initial_judgements)

                unique_groups = current.item.get_unique_record_groups()
                target_count = len(unique_groups)

                # Set group to assign
                await pilot.press("a")
                await pilot.pause()

                # Robust loop: Fill visible page, scroll right, repeat.
                iteration: int = 0
                while len(current.assignments) < target_count:
                    if iteration > 20:
                        pytest.fail("Failed to assign all columns in 20 iterations")

                    # Assign visible batch (1-9) to group 'a'
                    for key in "123456789":
                        await pilot.press(key)
                        await pilot.pause()

                    if len(current.assignments) == target_count:
                        break

                    await pilot.press("right")
                    await pilot.pause()
                    iteration += 1

                # Verify all columns are now assigned
                assert len(current.assignments) == target_count, (
                    "Not all columns assigned: "
                    f"{len(current.assignments)}/{target_count}\n"
                    f"Assignments: {sorted(current.assignments.keys())}\n"
                )

                # Now submit should succeed
                await pilot.press("space")
                await pilot.pause()

                final_judgements, _ = self.backend.get_judgements()
                assert len(final_judgements) == initial_judgement_count + 1, (
                    "Judgement was not saved after complete assignment"
                )

    @pytest.mark.asyncio
    async def test_app_widgets_are_visible(self) -> None:
        """Test that all app widgets are visible with non-zero dimensions."""
        with self.scenario(self.backend, "dedupe") as dag:
            model_name: str = "naive_test_crn"

            loaded_dag: DAG = dag.dag.load_pending().set_client(self.warehouse_engine)

            app = EntityResolutionApp(
                resolution=loaded_dag.get_model(model_name),
                num_samples=3,
                user="test_user",
                dag=loaded_dag,
            )

            async with app.run_test() as pilot:
                await pilot.pause()

                # Check table
                table = app.query_one(ComparisonDisplayTable)
                assert table is not None
                assert table.size.width > 0, "Table has no width - not rendering"
                assert table.size.height > 0, "Table has no height - not rendering"
                assert table.row_count > 0, "Table has no rows"
                assert len(table.ordered_columns) > 0, "Table has no columns"

                # Check status bar widgets
                current_group_label = app.query_one("#current-group-label", Label)
                assert current_group_label is not None
                assert current_group_label.size.width > 0, (
                    "Current group label not rendering"
                )
                assert current_group_label.size.height > 0

                status_right = app.query_one("#status-right", Label)
                assert status_right is not None
                assert status_right.size.width > 0, "Status right not rendering"
                assert status_right.size.height > 0

                # Check footer
                footer = app.query_one(Footer)
                assert footer is not None
                assert footer.size.width > 0, "Footer not rendering"
                assert footer.size.height > 0
